# Copyright 2021 Zilliz. All rights reserved.

#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import importlib
from typing import Dict, Tuple, List, Callable

import numpy as np
import torch
from torch import nn
from torch.optim import Optimizer
from tqdm import tqdm

from towhee.utils.log import trainer_log
from towhee.trainer.utils.trainer_utils import is_main_process
from towhee.trainer.utils.file_utils import is_tensorboard_available

__all__ = [
    "Callback",
    "CallbackList",
    "TrainerControl",
    "EarlyStoppingCallback",
    "ModelCheckpointCallback",
    "TensorBoardCallBack",
    "PrintCallBack",
    "ProgressBarCallBack"
]


def _get_summary_writer_constructor():
    if not is_tensorboard_available():
        trainer_log.info("can not import tensorboard.")
        return None
    try:
        tensorboard_module = importlib.import_module("torch.utils.tensorboard")
        summary_writer_constructor = tensorboard_module.SummaryWriter
        trainer_log.info("Use tensorboard. And please observe the logs in  http://localhost:6007/")
        return summary_writer_constructor
    except ImportError:
        trainer_log.info("can not import tensorboard.")
        return None


class TrainerControl:
    """
    `TrainerControl` defines a set of current control status which trainer
    can get and take the corresponding action. It can be used by customized
    `Callback` to interfere the trainer.

    Args:
        should_training_stop: (`bool`)
            whether or not training should be interrupted.
        should_epoch_stop: (`bool`)
            whether or not current training epoch should be interrupted.
        should_save: (`bool`)
            whether or not trainer should save current model.
        should_evaluate: (`bool`)
            whether or not trainer should evaluate current model.
        should_log: (`bool`)
            whether or not trainer should report the log.
    """

    def __init__(self,
                 should_training_stop: bool = False,
                 should_epoch_stop: bool = False,
                 should_save: bool = False,
                 should_evaluate=False,
                 should_log=False):
        self.should_training_stop = should_training_stop
        self.should_epoch_stop = should_epoch_stop
        self.should_save = should_save
        self.should_evaluate = should_evaluate
        self.should_log = should_log


class Callback:
    """
    `Callback` defines a set of functions which will be called in the training
    process. Customized `Callback` could inherent the base `Callback` and
    overwrite its methods to control the training process or handle the training
    information.
    """

    def __init__(self):
        self.model = None
        self.optimizer = None
        self.trainercontrol = None

    def set_model(self, model: nn.Module) -> None:
        """
        Set the model to callback.

        Args:
            model (`torch.nn.Module`):
                The model which callback can operate.
        """
        self.model = model

    def set_optimizer(self, optimizer: Optimizer) -> None:
        """
        Set the optimizer to callback.

        Args:
            optimizer (`torch.optim.Optimizer`):
                The optimizer which callback can operate.
        """
        self.optimizer = optimizer

    def set_trainercontrol(self, trainercontrol: TrainerControl) -> None:
        """
        Set the trainercontrol to callback.

        Args:
            trainercontrol (`towhee.trainer.callback.TrainerControl`):
                The trainercontrol which callback can operate.
        """
        self.trainercontrol = trainercontrol

    def on_batch_begin(self, batch: Tuple, logs: Dict) -> None:
        """
        Hook function invoked before every batch calculation.

        Args:
            batch (`Tuple`):
                The data batch to calculate.
            logs (`Dict`)
                Kv store to save and load info.
        """
        pass

    def on_batch_end(self, batch: Tuple, logs: Dict) -> None:
        """
        Hook function invoked after every batch calculation.

        Args:
            batch (`Tuple`):
                The data batch to calculate.
            logs (`Dict`)
                Kv store to save and load info.
        """
        pass

    def on_epoch_begin(self, epochs: int, logs: Dict) -> None:
        """
        Hook function invoked before each epoch.

        Args:
            epochs (`int`):
                Epoch index.
            logs (`Dict`):
                Kv store to save and load info.
        """
        pass

    def on_epoch_end(self, epochs: int, logs: Dict) -> None:
        """
        Hook function invoked after each epoch.

        Args:
            epochs (`int`):
                Epoch index.
            logs (`Dict`):
                Kv store to save and load info.
        """
        pass

    def on_train_begin(self, logs: Dict) -> None:
        """
        Hook function invoked before train stage.

        Args:
            logs (`Dict`):
               Kv store to save and load info.
        """
        pass

    def on_train_end(self, logs: Dict) -> None:
        """
        Hook function invoked after train stage.

        Args:
            logs (`Dict`):
               Kv store to save and load info.
        """
        pass

    def on_train_batch_begin(self, batch: Tuple, logs: Dict) -> None:
        """
        Hook function invoked before train stage.

        Args:
            logs (`Dict`):
               Kv store to save and load info.
        """
        self.on_batch_begin(batch, logs)

    def on_train_batch_end(self, batch: Tuple, logs: Dict) -> None:
        """
        Hook function invoked before every batch calculation in train stage.

        Args:
            batch (`Tuple`):
                The data batch to calculate.
            logs (`Dict`)
                Kv store to save and load info.
        """
        self.on_batch_end(batch, logs)

    def on_eval_batch_begin(self, batch: Tuple, logs: Dict) -> None:
        """
        Hook function invoked before every batch calculation in evaluate stage.

        Args:
            batch (`Tuple`):
                The data batch to calculate.
            logs (`Dict`)
                Kv store to save and load info.
        """
        self.on_batch_begin(batch, logs)

    def on_eval_batch_end(self, batch: Tuple, logs: Dict) -> None:
        """
        Hook function invoked after every batch calculation in evaluate stage.

        Args:
            batch (`Tuple`):
                The data batch to calculate.
            logs (`Dict`)
                Kv store to save and load info.
        """
        self.on_batch_end(batch, logs)

    def on_eval_begin(self, logs: Dict) -> None:
        """
        Hook function invoked before evaluate stage.

        Args:
            logs (`Dict`):
               Kv store to save and load info.
        """
        pass

    def on_eval_end(self, logs: Dict) -> None:
        """
        Hook function invoked after evaluate stage.

        Args:
            logs (`Dict`):
               Kv store to save and load info.
        """
        pass


class CallbackList:
    """
    `CallbackList` aggregate multiple `Callback` in the same object. Invoke the
    callbacks of `CallbackList` will invoke corresponding callback in each
    "Callback" in the FIFO sequential order.

    Args:
        callbacks (`List[towhee.trainer.callback.Callback]`)
            A list of callbacks which methods will be called simultaneously.

    Example:

    """

    def __init__(self, callbacks: List[Callback] = None):
        self.callbacks = []
        if callbacks is not None:
            self.callbacks.extend(callbacks)

    def __len__(self) -> int:
        return len(self.callbacks)

    def __getitem__(self, idx: int) -> Callback:
        return self.callbacks[idx]

    def __repr__(self):
        callback_desc = ""
        for cb in self.callbacks:
            callback_desc += cb.__repr__() + ","
        return "towhee.trainer.CallbackList([{}])".format(callback_desc)

    def set_model(self, model: nn.Module):
        """
        Set the model to callback.

        Args:
            model (`torch.nn.Module`):
                The model which callback can operate.
        """
        self.model = model
        for cb in self.callbacks:
            cb.set_model(model)

    def set_optimizer(self, optimizer: Optimizer):
        """
        Set the optimizer to callback.

        Args:
            optimizer (`torch.optim.Optimizer`):
                The optimizer which callback can operate.
        """
        self.optimizer = optimizer
        for cb in self.callbacks:
            cb.set_optimizer(optimizer)

    def set_trainercontrol(self, trainercontrol: TrainerControl):
        """
        Set the trainercontrol to callback.

        Args:
            trainercontrol (`towhee.trainer.callback.TrainerControl`):
                The trainercontrol which callback can operate.
        """
        self.trainercontrol = TrainerControl
        for cb in self.callbacks:
            cb.set_trainercontrol(trainercontrol)

    def add_callback(self, callback: Callback, singleton: bool = True):
        """
        Args:
            callback (`towhee.trainer.callback.Callback`):
                The callback need to be added.
            singleton (`bool`):
                If set true, only one instance of same `Callback` will remain in callbacklist.
        """
        if singleton:
            for old_callback in self.callbacks:
                if old_callback.__class__.__name__ == callback.__class__.__name__:
                    self.callbacks.remove(old_callback)
        self.callbacks.append(callback)

    def pop_callback(self, callback: Callback):
        """
        Args:
            callback (`towhee.trainer.callback.Callback`)
                The callback need to be removed from callback list.
        """
        if callback in self.callbacks:
            self.callbacks.remove(callback)

    def on_batch_begin(self, batch: Tuple, logs: Dict) -> None:
        """
        Hook function invoked before every batch calculation.

        Args:
            batch (`Tuple`):
                The data batch to calculate.
            logs (`Dict`)
                Kv store to save and load info.
        """
        for cb in self.callbacks:
            cb.on_batch_begin(batch, logs)

    def on_batch_end(self, batch: Tuple, logs: Dict) -> None:
        """
        Hook function invoked after every batch calculation.

        Args:
            batch (`Tuple`):
                The data batch to calculate.
            logs (`Dict`)
                Kv store to save and load info.
        """
        for cb in self.callbacks:
            cb.on_batch_end(batch, logs)

    def on_epoch_begin(self, epochs: int, logs: Dict) -> None:
        """
        Hook function invoked before each epoch.

        Args:
            epochs (`int`):
                Epoch index.
            logs (`Dict`):
                Kv store to save and load info.
        """
        for cb in self.callbacks:
            cb.on_epoch_begin(epochs, logs)

    def on_epoch_end(self, epochs: int, logs: Dict) -> None:
        """
        Hook function invoked after each epoch.

        Args:
            epochs (`int`):
                Epoch index.
            logs (`Dict`):
                Kv store to save and load info.
        """
        for cb in self.callbacks:
            cb.on_epoch_end(epochs, logs)

    def on_train_begin(self, logs: Dict) -> None:
        """
        Hook function invoked before train stage.

        Args:
            logs (`Dict`):
               Kv store to save and load info.
        """
        for cb in self.callbacks:
            cb.on_train_begin(logs)

    def on_train_end(self, logs: Dict) -> None:
        """
        Hook function invoked after train stage.

        Args:
            logs (`Dict`):
               Kv store to save and load info.
        """
        for cb in self.callbacks:
            cb.on_train_end(logs)

    def on_train_batch_begin(self, batch: Tuple, logs: Dict) -> None:
        """
        Hook function invoked before train stage.

        Args:
            logs (`Dict`):
               Kv store to save and load info.
        """
        for cb in self.callbacks:
            cb.on_train_batch_begin(batch, logs)

    def on_train_batch_end(self, batch: Tuple, logs: Dict) -> None:
        """
        Hook function invoked before every batch calculation in train stage.

        Args:
            batch (`Tuple`):
                The data batch to calculate.
            logs (`Dict`)
                Kv store to save and load info.
        """
        for cb in self.callbacks:
            cb.on_train_batch_end(batch, logs)

    def on_eval_batch_begin(self, batch: Tuple, logs: Dict) -> None:
        """
        Hook function invoked before every batch calculation in evaluate stage.

        Args:
            batch (`Tuple`):
                The data batch to calculate.
            logs (`Dict`)
                Kv store to save and load info.
        """
        for cb in self.callbacks:
            cb.on_eval_batch_begin(batch, logs)

    def on_eval_batch_end(self, batch: Tuple, logs: Dict) -> None:
        """
        Hook function invoked after every batch calculation in evaluate stage.

        Args:
            batch (`Tuple`):
                The data batch to calculate.
            logs (`Dict`)
                Kv store to save and load info.
        """
        for cb in self.callbacks:
            cb.on_eval_batch_end(batch, logs)

    def on_eval_begin(self, logs: Dict) -> None:
        """
        Hook function invoked before evaluate stage.

        Args:
            logs (`Dict`):
               Kv store to save and load info.
        """
        for cb in self.callbacks:
            cb.on_eval_begin(logs)

    def on_eval_end(self, logs: Dict) -> None:
        """
        Hook function invoked after evaluate stage.

        Args:
            logs (`Dict`):
               Kv store to save and load info.
        """
        for cb in self.callbacks:
            cb.on_eval_end(logs)


class EarlyStoppingCallback(Callback):
    """
    EarlyStoppingCallback

    Assuming the goal of a training is to minimize the loss. With this, the
    metric to be monitored would be `'loss'`, and mode would be `'min'`.
    Training loop will check at end of every epoch whether the loss is no
    longer decreasing, considering the `min_delta` and `patience` if
    applicable. Once it's found no longer decreasing. `trainercontrol.
    should_training_stop` is marked True.

    Args:
        trainercontrol (`towhee.trainer.callback.TrainerControl`):
            The trainercontrol which callback can operate.
        monitor (`str`):
            Quantity to be monitored.
        min_delta (`float`):
            Minimum change in the monitored quantity to qualify as an
            improvement, i.e. an absolute change of less than min_delta,
            will count as no improvement.
        patience (`str`):
            Number of epochs with no improvement after which training will
            be stopped.
        mode (`str`):
            One of `{"min", "max"}`. In `min` mode, training will stop when
            the quantity monitored has stopped decreasing; in `"max"`
            mode it will stop when the quantity monitored has stopped increasing.
        baseline (`float`):
            Baseline value for the monitored quantity.
            Training will stop if the model doesn't show improvement over the
            baseline.
    """

    def __init__(self,
                 trainercontrol: TrainerControl,
                 monitor: str,
                 min_delta: float = 0,
                 patience: int = 0,
                 mode: str = "max",
                 baseline: float = None
                 ):
        super(EarlyStoppingCallback).__init__()
        self.trainercontrol = trainercontrol
        self.monitor = monitor
        self.patience = patience
        self.baseline = baseline
        self.min_delta = abs(min_delta)
        self.wait = 0
        self.stopped_epoch = 0
        self.best_weights = None

        assert mode in ["max", "min"]

        if mode == "min":
            self.monitor_op = np.less
        else:
            self.monitor_op = np.greater

        if self.monitor_op == np.greater:
            self.min_delta *= 1
        else:
            self.min_delta *= -1

    def on_train_begin(self, logs: Dict = None):
        self.wait = 0
        self.stopped_epoch = 0
        self.best = np.Inf if self.monitor_op == np.less else -np.Inf
        self.best_weights = None
        self.best_epoch = 0

    def on_epoch_end(self, epochs: int, logs: Dict = None):
        current = self.get_monitor_value(logs)
        if current is None:
            return

        self.wait += 1
        if self._is_improvement(current, self.best):
            self.best = current
            self.best_epoch = epochs
            if self.baseline is None or self._is_improvement(current, self.baseline):
                self.wait = 0

        if self.wait >= self.patience and epochs > 0 and current != 0:
            self.stopped_epoch = epochs
            self.trainercontrol.should_training_stop = True

    def on_train_end(self, logs: Dict = None):
        if self.stopped_epoch > 0:
            if is_main_process():
                trainer_log.warning(
                    "monitoring %s not be better then %s on epoch %s for waiting for %s epochs. Early stop on epoch %s.",
                    self.monitor, self.best, self.best_epoch, self.wait, self.stopped_epoch)

    def get_monitor_value(self, logs: Dict):
        logs = logs or {}
        monitor_value = logs.get(self.monitor)

        return monitor_value

    def _is_improvement(self, monitor_value: float, reference_value: float):
        return self.monitor_op(monitor_value - self.min_delta, reference_value)


class ModelCheckpointCallback(Callback):
    """
    ModelCheckpointCallback is intended to save the model at some interval. It can
    be set in epoch mode or iteration mode. Only one of `every_n_epoch` and
    `every_n_iteration` can be set to a positive value and the `trainer.should_save`
    will set to True when the condion meets.

    Args:
        trainercontrol (`TrainerControl`):
            The trainercontrol which callback can operate.
        filepath (`str`):
            Filepath to save the model.
        every_n_epoch (`int`):
            Save the model after n epochs.
        every_n_iteration (`int`):
            Save the model after n iterations.
    """

    def __init__(self,
                 trainercontrol: TrainerControl,
                 filepath: str = "./",
                 every_n_epoch: int = -1,
                 every_n_iteration: int = -1):
        super(ModelCheckpointCallback).__init__()
        self.trainercontrol = trainercontrol
        if every_n_epoch != -1:
            assert every_n_iteration == -1
        self.every_n_epoch = every_n_epoch
        if every_n_iteration != -1:
            assert every_n_epoch == -1
        self.every_n_iteration = every_n_iteration

        self.save_path_prefix = filepath
        self.n_iteration = 0
        assert (self.every_n_epoch != 0 and self.every_n_epoch > -2)
        assert (self.every_n_iteration != 0 and self.every_n_iteration > -2)

    def on_epoch_end(self, epochs: int, logs: Dict = None):
        if self.every_n_epoch == -1:
            return
        if self.trainercontrol.should_save is True:
            self.trainercontrol.should_save = False
        if self.every_n_epoch >= 1 and epochs % self.every_n_epoch == 0:
            self._save_model()

    def on_batch_end(self, batch: Tuple, logs: Dict = None):
        if self.every_n_iteration == -1:
            return
        if self.trainercontrol.should_save is True:
            self.trainercontrol.should_save = False
        if self.every_n_iteration >= 1 and self.n_iteration % self.every_n_iteration == 0:
            self._save_model()
        self.n_iteration += 1

    def _save_model(self):
        self.trainercontrol.should_save = True


class TensorBoardCallBack(Callback):
    """
    TensorBoardCallBack is intended to record the essential value(e.g. epoch_loss)
    to tensorboard after each iteration. If tensorboard is available, you can see
    the tensorboard in localhost:6006.

    Args:
        summary_writer_constructor (`Callable`):
            Function which construct tensorboard summary writer.
        log_dir (`str`):
            Save directory location.
        comment (`str`):
            Comment log_dir suffix appended to the default log_dir.
    """

    def __init__(self, summary_writer_constructor: Callable, log_dir: str = None, comment: str = ""):
        super().__init__()
        self.tb_writer = summary_writer_constructor(log_dir, comment=comment)

    def on_train_batch_end(self, batch: Tuple, logs: Dict) -> None:
        global_step = logs["global_step"]
        step_loss = logs["step_loss"]
        epoch_loss = logs["epoch_loss"]
        epoch_metric = logs["epoch_metric"]
        lr = logs["lr"]
        if is_main_process():
            self.tb_writer.add_scalar("lr", lr, global_step)
            self.tb_writer.add_scalar("epoch_loss", epoch_loss, global_step)
            self.tb_writer.add_scalar("step_loss", step_loss, global_step)
            self.tb_writer.add_scalar("epoch_metric", epoch_metric, global_step)

    def on_eval_batch_end(self, batch: Tuple, logs: Dict) -> None:
        eval_global_step = logs["eval_global_step"]
        eval_step_loss = logs["eval_step_loss"]
        eval_epoch_loss = logs["eval_epoch_loss"]
        eval_epoch_metric = logs["eval_epoch_metric"]
        if is_main_process():
            self.tb_writer.add_scalar("eval_step_loss", eval_step_loss, eval_global_step)
            self.tb_writer.add_scalar("eval_epoch_loss", eval_epoch_loss, eval_global_step)
            self.tb_writer.add_scalar("eval_epoch_metric", eval_epoch_metric, eval_global_step)


class PrintCallBack(Callback):
    """
    PrintCallBack is intended to print logs on the screen.

    Args:
        total_epoch_num (`int`):
            Epoch numbers expected to run.
        step_frequency (`int`):
            Print information in every n steps.
    """

    def __init__(self, total_epoch_num: int, step_frequency: int = 16):
        super().__init__()
        self.step_frequency = step_frequency
        self.total_epoch_num = total_epoch_num

    def on_train_batch_end(self, batch: Tuple, logs: Dict) -> None:
        if is_main_process():
            global_step = logs["global_step"]
            if global_step % self.step_frequency == 0:
                print("epoch={}/{}, global_step={}, epoch_loss={}, epoch_metric={}"
                      .format(logs["epoch"], self.total_epoch_num,
                              global_step,
                              logs["epoch_loss"],
                              logs["epoch_metric"]))

    def on_eval_batch_end(self, batch: Tuple, logs: Dict) -> None:
        if is_main_process():
            eval_global_step = logs["eval_global_step"]
            if eval_global_step % self.step_frequency == 0:
                print("epoch={}/{}, eval_global_step={}, eval_epoch_loss={}, eval_epoch_metric={}"
                      .format(logs["epoch"], self.total_epoch_num,
                              eval_global_step,
                              logs["eval_epoch_loss"],
                              logs["eval_epoch_metric"]))


class ProgressBarCallBack(Callback):
    """
    ProgressBarCallBack is intended to print a progress bar to visualize current
    training progress. The tqdm is used as the progress bar backend.

    Args:
        total_epoch_num (`int`):
            Epoch numbers expected to run.
        train_dataloader (`torch.utils.data.DataLoader`):
            training dataloader for tqdm to warp.
    """

    def __init__(self, total_epoch_num: int, train_dataloader: torch.utils.data.DataLoader):
        super().__init__()
        self.total_epoch_num = total_epoch_num
        self.raw_train_dataloader = train_dataloader
        self.now_tqdm_train_dataloader: tqdm = train_dataloader
        self.description = ""

    def on_train_batch_end(self, batch: Tuple, logs: Dict) -> None:
        if is_main_process():
            self.now_tqdm_train_dataloader.update(1)
            self.description = "[epoch {}/{}] loss={}, metric={}".format(logs["epoch"],
                                                                         int(self.total_epoch_num),
                                                                         round(logs["epoch_loss"], 3),
                                                                         round(logs["epoch_metric"], 3))
            self.now_tqdm_train_dataloader.set_description(self.description)

    def on_epoch_begin(self, epochs: int, logs: Dict) -> None:
        if is_main_process():
            self.now_tqdm_train_dataloader = None
            self.now_tqdm_train_dataloader = tqdm(self.raw_train_dataloader,
                                                  total=len(self.raw_train_dataloader),
                                                  unit="step")  # , file=sys.stdout)

    def on_eval_batch_end(self, batch: Tuple, logs: Dict) -> None:
        if is_main_process():
            self.description = "[epoch {}/{}] loss={}, metric={}, eval_loss={}, eval_metric={}".format(
                logs["epoch"],
                int(self.total_epoch_num),
                round(logs["epoch_loss"], 3),
                round(logs["epoch_metric"], 3), round(logs["eval_epoch_loss"], 3), round(logs["eval_epoch_metric"], 3))
            self.now_tqdm_train_dataloader.set_description(self.description)



# Copyright clcarwin. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.

import torch
import torch.nn.functional as F

from torch import nn
from torch.autograd import Variable


class FocalLoss(nn.Module):
    """
    Focal loss for object detection.
    Args:
        gamma (`float`): hyper-parameter of focal loss
        alpha (`float`): hyper-parameter of focal loss
        size_average (`bool`): whether to average the loss
    """
    def __init__(self, gamma: float = 0, alpha: float = None, size_average: bool = True):
        super().__init__()
        self.gamma = gamma
        self.alpha = alpha
        if isinstance(alpha, (float, int)):
            self.alpha = torch.Tensor([alpha, 1 - alpha])
        if isinstance(alpha, list):
            self.alpha = torch.Tensor(alpha)
        self.size_average = size_average

    def forward(self, input_s: torch.Tensor, target: torch.Tensor):
        """
        Args:
            input_s (`float`): predicted result
            target (`float`): ground truth
        """
        if input_s.dim() > 2:
            input_s = input_s.view(input_s.size(0), input_s.size(1), -1)  # N,C,H,W => N,C,H*W
            input_s = input_s.transpose(1, 2)  # N,C,H*W => N,H*W,C
            input_s = input_s.contiguous().view(-1, input_s.size(2))  # N,H*W,C => N*H*W,C
        target = target.view(-1, 1)

        logpt = F.log_softmax(input_s)
        logpt = logpt.gather(1, target)
        logpt = logpt.view(-1)
        pt = Variable(logpt.data.exp())

        if self.alpha is not None:
            if self.alpha.type() != input_s.data.type():
                self.alpha = self.alpha.type_as(input_s.data)
            at = self.alpha.gather(0, target.data.view(-1))
            logpt = logpt * Variable(at)

        loss = -1 * (1 - pt) ** self.gamma * logpt
        if self.size_average:
            return loss.mean()
        else:
            return loss.sum()



# original code from https://github.com/krasserm/perceiver-io


from torch import nn


class Sequential(nn.Sequential):
    """
    Sequential module for Perceiver https://arxiv.org/pdf/2103.03206.pdf.
    """
    def forward(self, *inputs):
        for module in self:
            if type(inputs) == tuple:
                inputs = module(*inputs)
            else:
                inputs = module(inputs)
        return inputs



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from copy import deepcopy

from towhee.engine.operator_runner.runner_base import RunnerBase
from towhee.errors import OpIOTypeError
from towhee.types._frame import FRAME


class FlatMapRunner(RunnerBase):
    """
    FlatMap, one input multiple outputs.
    """

    def _set_outputs(self, output: List[any]):
        if not isinstance(output, list):
            raise OpIOTypeError("Flatmap operator's output must be a list, not a {}".format(type(output)))

        for data in output:

            frame = deepcopy(self._frame)
            if frame.parent_path == '':
                frame.parent_path = str(frame.prev_id)
            else:
                frame.parent_path = '-'.join([frame.parent_path, str(frame.prev_id)])
            frame.prev_id = -1

            item = data._asdict()
            item.update({FRAME: frame})
            self._writer.write(item)



# Copyright 2021 biubug6 . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.

#adapted from https://github.com/biubug6/Pytorch_Retinaface
from typing import List

import torch
import torch.nn.functional as F
from torch import nn

from towhee.models.retina_face.utils import conv_bn1x1, conv_bn

class RetinaFaceFPN(nn.Module):
    """
    RetinaFaceFPN

    Feature Pyramid Network for RetinaFace.

    Args:
        in_channels_list (`List[int]`):
            input channel numbers ofr each FPN layer.
        out_channels (`int`):
            output channel number.
    """
    def __init__(self,in_channels_list: List[int],out_channels:int):
        super().__init__()
        leaky = 0
        if out_channels <= 64:
            leaky = 0.1
        self.output1 = conv_bn1x1(in_channels_list[0], out_channels, stride = 1, leaky = leaky)
        self.output2 = conv_bn1x1(in_channels_list[1], out_channels, stride = 1, leaky = leaky)
        self.output3 = conv_bn1x1(in_channels_list[2], out_channels, stride = 1, leaky = leaky)

        self.merge1 = conv_bn(out_channels, out_channels, leaky = leaky)
        self.merge2 = conv_bn(out_channels, out_channels, leaky = leaky)

    def forward(self, x: torch.FloatTensor):
        # names = list(input.keys())
        x = list(x.values())

        output1 = self.output1(x[0])
        output2 = self.output2(x[1])
        output3 = self.output3(x[2])

        up3 = F.interpolate(output3, size=[output2.size(2), output2.size(3)], mode="nearest")
        output2 = output2 + up3
        output2 = self.merge2(output2)

        up2 = F.interpolate(output2, size=[output1.size(2), output1.size(3)], mode="nearest")
        output1 = output1 + up2
        output1 = self.merge1(output1)

        out = [output1, output2, output3]
        return out



# coding=utf-8
# Copyright 2020-present the HuggingFace Inc. team and 2021 Zilliz.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utilities for the Trainer.
"""

import random
from typing import NamedTuple, Any, Generator, Callable, Dict, Union
from collections.abc import Mapping
from enum import Enum
import numpy as np
import torch
import os
from pathlib import Path
import torch.distributed as dist
from torch import nn

NAME = "name_"


def set_seed(seed: int):
    """
    Helper function for reproducible behavior to set the seed in ``random``, ``numpy``, ``torch``.

    Args:
        seed (:obj:`int`): The seed to set.
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)


def get_last_checkpoint(out_dir: str):
    checkpoints = [path for path in Path(out_dir).iterdir() if path.is_dir()]
    if len(checkpoints) == 0:
        raise ValueError(f"No checkpoints found at {out_dir}.")
    return max(checkpoints, key=os.path.getmtime).resolve()


class TrainOutput(NamedTuple):
    global_step: int
    training_loss: float


STATE_CHECKPOINT_NAME = "state.pt"
MODEL_NAME = "model.pth"


# _re_checkpoint = re.compile(r"^" + CHECKPOINT_NAME + r"\-(\d+)$")

class SchedulerType(Enum):
    LINEAR = "linear"
    COSINE = "cosine"
    COSINE_WITH_RESTARTS = "cosine_with_restarts"
    POLYNOMIAL = "polynomial"
    CONSTANT = "constant"
    CONSTANT_WITH_WARMUP = "constant_with_warmup"


def is_dist_avail_and_initialized():
    if not dist.is_available():
        return False
    if not dist.is_initialized():
        return False
    return True


def get_world_size():
    if not is_dist_avail_and_initialized():
        return 1
    return dist.get_world_size()


def reduce_value(value, average=True):
    world_size = get_world_size()
    if world_size < 2:  # one gpu
        return value
    with torch.no_grad():
        dist.all_reduce(value)
        if average:
            value /= world_size
        return value


def get_rank():
    if not is_dist_avail_and_initialized():
        return 0
    return dist.get_rank()


def is_main_process():
    return get_rank() == 0


def honor_type(obj, generator: Generator):
    """
    Cast a generator to the same type as obj (list, tuple or namedtuple)
    """
    # There is no direct check whether an object if of type namedtuple sadly, this is a workaround.
    if isinstance(obj, tuple) and hasattr(obj, "_fields"):
        # Can instantiate a namedtuple from a generator directly, contrary to a tuple/list.
        return type(obj)(*list(generator))
    return type(obj)(generator)


def is_torch_tensor(tensor: Any):
    return isinstance(tensor, torch.Tensor)


def recursively_apply(func: Callable, data: Any, *args, test_type: Callable = is_torch_tensor,
                      error_on_other_type: bool = False, **kwargs):
    """
    Recursively apply a function on a data structure that is a nested list/tuple/dictionary of a given base type.

    Args:
        func (:obj:`callable`):
            The function to recursively apply.
        data (nested list/tuple/dictionary of :obj:`main_type`):
            The data on which to apply :obj:`func`
        *args:
            Positional arguments that will be passed to :obj:`func` when applied on the unpacked data.
        main_type (:obj:`type`, `optional`, defaults to :obj:`torch.Tensor`):
            The base type of the objects to which apply :obj:`func`.
        error_on_other_type (:obj:`bool`, `optional`, defaults to :obj:`False`):
            Whether to return an error or not if after unpacking :obj:`data`, we get on an object that is not of type
            :obj:`main_type`. If :obj:`False`, the function will leave objects of types different than :obj:`main_type`
            unchanged.
        **kwargs:
            Keyword arguments that will be passed to :obj:`func` when applied on the unpacked data.

    Returns:
        The same data structure as :obj:`data` with :obj:`func` applied to every object of type :obj:`main_type`.
    """
    if isinstance(data, (tuple, list)):
        return honor_type(
            data,
            (
                recursively_apply(
                    func, o, *args, test_type=test_type, error_on_other_type=error_on_other_type, **kwargs
                )
                for o in data
            ),
        )
    elif isinstance(data, Mapping):
        return type(data)(
            {
                k: recursively_apply(
                    func, v, *args, test_type=test_type, error_on_other_type=error_on_other_type, **kwargs
                )
                for k, v in data.items()
            }
        )
    elif test_type(data):
        return func(data, *args, **kwargs)
    elif error_on_other_type:
        raise TypeError(
            f"Can't apply {func.__name__} on object of type {type(data)}, only of nested list/tuple/dicts of objects "
            f"that satisfy {test_type.__name__}."
        )
    return data


def send_to_device(tensor: Any, device: torch.device):
    """
    Recursively sends the elements in a nested list/tuple/dictionary of tensors to a given device.
    Borrowed from huggingface/accelerate.
    Args:
        tensor (nested list/tuple/dictionary of :obj:`torch.Tensor`):
            The data to send to a given device.
        device (:obj:`torch.device`):
            The device to send the data to

    Returns:
        The same data structure as :obj:`tensor` with all tensors sent to the proper device.
    """

    def _send_to_device(t, device):
        return t.to(device)

    def _has_to_method(t):
        return hasattr(t, "to")

    return recursively_apply(_send_to_device, tensor, device, test_type=_has_to_method)


def unwrap_model(model: nn.Module) -> nn.Module:
    """
    Unwraps a model from potential containers (as used in distributed training).

    Args:
        model (`torch.nn.Module`): The model to unwrap.
    """
    # since there could be multiple levels of wrapping, unwrap recursively
    if hasattr(model, "module"):
        return model.module
    else:
        return model


def _construct_loss_from_config(module: Any, config: Union[str, Dict]):
    """
    construct from the config, the config can be class name as a `str`, or a dict containing the construct parameters.
    """
    instance = None
    if isinstance(config, str):
        construct_name = getattr(module, config)
        instance = construct_name()
    elif isinstance(config, Dict):
        optimizer_construct_name = config[NAME]
        construct_name = getattr(module, optimizer_construct_name)
        kwargs = {}
        for arg_name in config:
            if arg_name != NAME:
                kwargs[arg_name] = config[arg_name]
        instance = construct_name(**kwargs)
    return instance


def _construct_scheduler_from_config(optimizer: torch.optim.Optimizer, module: Any, config: Union[str, Dict]):
    """
    construct from the config, the config can be class name as a `str`, or a dict containing the construct parameters.
    """
    instance = None
    if isinstance(config, str):
        construct_name = getattr(module, config)
        instance = construct_name(optimizer)
    elif isinstance(config, Dict):
        scheduler_construct_name = config[NAME]
        construct_name = getattr(module, scheduler_construct_name)
        kwargs = {}
        for arg_name in config:
            if arg_name != NAME:
                kwargs[arg_name] = config[arg_name]
        instance = construct_name(optimizer, **kwargs)
    return instance


def _construct_optimizer_from_config(module: Any, config: Union[str, Dict], model=None):
    """
    construct from the config, the config can be class name as a `str`, or a dict containing the construct parameters.
    """
    instance = None
    if isinstance(config, str):
        construct_name = getattr(module, config)
        if model is not None:
            trainable_params = [p for p in model.parameters() if p.requires_grad]
            instance = construct_name(trainable_params)
    elif isinstance(config, Dict):
        optimizer_construct_name = config[NAME]
        construct_name = getattr(module, optimizer_construct_name)
        kwargs = {}
        for arg_name in config:
            if arg_name != NAME:
                kwargs[arg_name] = config[arg_name]
        if model is not None:
            trainable_params = [p for p in model.parameters() if p.requires_grad]
            instance = construct_name(trainable_params, **kwargs)
    return instance



# original code from https://github.com/krasserm/perceiver-io
# modified by Zilliz


from towhee.models.perceiver.create_self_attention import create_self_attention
from towhee.models.perceiver.Sequential import Sequential


def create_self_attention_block(
    num_layers: int, num_channels: int, num_heads: int, dropout: float, activation_checkpoint: bool = False
):
    """
    Self attention block for Perceiver https://arxiv.org/pdf/2103.03206.pdf.
    Args:
        num_layers (`int`):
            Number of layers.
        num_channels (`int`):
            Number of channels.
        num_heads (`int`):
            Number of parallel attention heads.
        dropout (`nn.Module`):
            Dropout probability.
        activation_checkpoint (`bool`):
            Use activation checkpointing.
    """
    layers = [create_self_attention(num_channels, num_heads, dropout, activation_checkpoint)
              for _ in range(num_layers)]
    return Sequential(*layers)



# Copyright 2021 Microsoft . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.
from typing import Callable
from itertools import repeat
import collections.abc

# From PyTorch internals
def _ntuple(n: int) -> Callable:
    def parse(x):
        if isinstance(x, collections.abc.Iterable):
            return x
        return tuple(repeat(x, n))
    return parse

to_1tuple = _ntuple(1)
to_2tuple = _ntuple(2)
to_3tuple = _ntuple(3)
to_4tuple = _ntuple(4)
to_ntuple = _ntuple

def make_divisible(v: int, divisor: int=8, min_value: int=None, round_limit: float=.9) -> int:
    min_value = min_value or divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < round_limit * v:
        new_v += divisor
    return new_v



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union
from pathlib import PosixPath

from towhee.types import Image
from towhee.utils.log import engine_log
from towhee.utils.ndarray_utils import from_src

try:
    from matplotlib import pyplot
except ModuleNotFoundError as e:
    engine_log.error('matplotlib not found, you can install via `pip install matplotlib`.')
    raise ModuleNotFoundError('matplotlib not found, you can install via `pip install matplotlib`.') from e


def plot_img(img: Union[str, PosixPath, Image, list], title: str = None):
    """
    Plot one image in notebook.

    Args:
        img (`Union[str, PosixPath, Image]`):
            Image to plot.
        title (`str`):
            Title of the image to show.
    """
    if isinstance(img, list):
        plot_img_list(img, title)
    else:
        if isinstance(img, Union[str, PosixPath].__args__):
            img = from_src(img)
        _, ax = pyplot.subplots(1, 1, figsize=(5, 5))
        try:
            ax.imshow(img.to_ndarray())
        except: # pylint: disable=bare-except
            ax.imshow(img.cv2_to_rgb())
        ax.axis('off')
        if title:
            ax.set_title(f'{title:.04f}\n')


def plot_img_list(data: list, titles: list = None):
    """
    Plot image list in notebook.

    Args:
        data (`list`):
            All the dataset.
        titles (`list`):
            Title list of the image to show.
    """
    if len(data) == 1:
        return plot_img(data[0], titles)
    _, axs = pyplot.subplots(1, len(data), figsize=(5 * len(data), 5 * len(data)))
    for ax, i in zip(axs, range(len(data))):
        img = data[i]
        if isinstance(img, Union[str, PosixPath].__args__):
            img = from_src(img)
        try:
            ax.imshow(img.to_ndarray())
        except: # pylint: disable=bare-except
            ax.imshow(img.cv2_to_rgb())
        ax.axis('off')
        if titles and titles[i]:
            ax.set_title(f'{titles[i]:.04f}\n')



# Original pytorch implementation by:
# 'An Image Is Worth 16 x 16 Words: Transformers for Image Recognition at Scale'
#       - https://arxiv.org/abs/2010.11929
# 'How to train your ViT? Data, Augmentation, and Regularization in Vision Transformers'
#       - https://arxiv.org/abs/2106.10270
#
# Original code by / Copyright 2020, Ross Wightman.
# Modifications & additions by / Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from torch import nn

class MultiHeadAttention(nn.Module):
    """
    Multi-head attention layer.
    Args:
        dim (int): number of features
        num_heads (int): number of heads
        qkv_bias (bool): if add bias to qkv layer
        qk_scale (float): number to scale qk
        attn_drop_ratio (float): drop rate of attention layer
        proj_drop_ratio (float): drop rate of projection layer
    """
    def __init__(self,
                 dim,
                 num_heads=8,
                 qkv_bias=False,
                 qk_scale=None,
                 attn_drop_ratio=0,
                 proj_drop_ratio=0):
        super().__init__()
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.scale = qk_scale or self.head_dim ** -0.5
        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop_ratio)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop_ratio)

    def forward(self, x):
        batch_size, new_num_patch, dim = x.shape

        qkv = self.qkv(x).reshape(
            batch_size,
            new_num_patch,
            3,
            self.num_heads,
            self.head_dim,
        ).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(batch_size, new_num_patch, dim)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x

# if __name__=='__main__':
#     import torch
#
#     input_x = torch.rand(1, 196+1, 768) # shape of output from patch_embed
#     model = MultiHeadAttention(dim=768)
#     out = model.forward(input_x)
#
#     print(out.shape)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
import threading
from collections import namedtuple
from typing import Dict, Tuple, Union, List

from towhee.dataframe.dataframe import DataFrame
from towhee.dataframe.iterators import DataFrameIterator, MapIterator, BatchIterator, WindowIterator


class ReaderBase(ABC):
    """
    The reader base class.
    The read() could be blocking or non-blocking function, if it's a blocking function,
    the runner may be blocked. When need to stop the graph, we call close to interrupting it.
    """

    @abstractmethod
    def read(self):
        raise NotImplementedError

    @abstractmethod
    def close(self):
        raise NotImplementedError


class DataFrameReader(ReaderBase):
    """
    Read data from input dataframes, unpack and combine data.
    One op_ctx has one dataframe reader.
    """

    def __init__(self, it: DataFrameIterator, op_inputs_index: Dict[str, int]):
        self._op_inputs_index = op_inputs_index
        self._iter = it

    @abstractmethod
    def read(self) -> Union[Dict[str, any], List[Dict[str, any]]]:
        pass

    @property
    def size(self) -> int:
        return self._iter.accessible_size

    @abstractmethod
    def close(self):
        raise NotImplementedError

    def _to_op_inputs(self, cols: Tuple) -> Dict[str, any]:
        """
        Read from cols, combine op inputs
        """
        ret = {}
        for key, index in self._op_inputs_index.items():
            ret[key] = cols[index]
        return ret


class BlockMapReaderWithOriginData(DataFrameReader):
    """
    Return both op's input data and origin data.
    """

    def __init__(
        self,
        input_df: DataFrame,
        op_inputs_index: Dict[str, int]
    ):
        super().__init__(MapIterator(input_df, True), op_inputs_index)
        self._lock = threading.Lock()
        self._close = False

    def read(self) -> Tuple[Dict[str, any], Tuple]:
        """
        Read data from dataframe, get cols by operator_repr info
        """
        if self._close:
            raise StopIteration

        with self._lock:
            data = next(self._iter)
            # map and batch iterators both return list of tuples
            data = data[0]
            if self._close:
                raise StopIteration

            if not data:
                return {}, ()
            return self._to_op_inputs(data), data

    def close(self):
        self._close = True
        self._iter.notify()


class BatchFrameReader(DataFrameReader):
    """
    Batch reader.
    """

    def __init__(self, input_df: DataFrame, op_inputs_index: Dict[str, int],
                 batch_size: int, step: int):
        assert batch_size >= 1 and step >= 1
        super().__init__(BatchIterator(input_df, batch_size, step, True), op_inputs_index)
        self._close = False
        self._lock = threading.Lock()

    def read(self) -> List[Dict[str, any]]:
        if self._close:
            raise StopIteration

        with self._lock:
            data = next(self._iter)
            if self._close:
                raise StopIteration

            if not data:
                return []
            else:
                res = []
                for row in data:
                    data_dict = self._to_op_inputs(row)
                    res.append(namedtuple('input', data_dict.keys())(**data_dict))
                return res

    def close(self):
        self._close = True
        self._iter.notify()


class TimeWindowReader(DataFrameReader):
    """
    Time window reader
    """

    def __init__(
        self,
        input_df: DataFrame,
        op_inputs_index: Dict[str, int],
        time_range_sec: int,
        time_step_sec: int
    ):
        super().__init__(WindowIterator(input_df, start = 0,
                                                window_size = time_range_sec,
                                                step = time_step_sec,
                                                use_timestamp = True,
                                                block = True), op_inputs_index)
        self._lock = threading.Lock()
        self._close = False

    def _format_to_namedtuple(self, rows):
        ret = []
        for row in rows:
            data_dict = self._to_op_inputs(row)
            ret.append(namedtuple('input', data_dict.keys())(**data_dict))
        return ret

    def read(self) -> List[Dict[str, any]]:
        if self._close:
            raise StopIteration

        with self._lock:
            data = next(self._iter)
            if self._close:
                raise StopIteration

            if not data:
                return []
            else:
                res = []
                for row in data:
                    data_dict = self._to_op_inputs(row)
                    res.append(namedtuple('input', data_dict.keys())(**data_dict))
                return res

    def close(self):
        self._close = True
        self._iter.notify()



# original code from https://github.com/krasserm/perceiver-io
# modified by Zilliz


from torch import nn

from towhee.models.perceiver.MultiHeadAttention import MultiHeadAttention


class SelfAttention(nn.Module):
    """
    Self attention for Perceiver https://arxiv.org/pdf/2103.03206.pdf.
    Args:
        num_channels (`int`):
            Number of channels.
        num_heads (`int`):
            Number of parallel attention heads.
        dropout (`nn.Module`):
            Dropout probability.
    """
    def __init__(self, num_channels: int, num_heads: int, dropout: float):
        super().__init__()
        self.norm = nn.LayerNorm(num_channels)
        self.attention = MultiHeadAttention(
            num_q_channels=num_channels, num_kv_channels=num_channels, num_heads=num_heads, dropout=dropout
        )

    def forward(self, x, pad_mask=None, attn_mask=None):
        """
        Forward function.
        Args:
            x (`Tensor`):
                Input tensor
            pad_mask (`Tensor`):
                Padding mask.
            attn_mask (`Tensor`):
                Attention mask.
        """
        x = self.norm(x)
        return self.attention(x, x, pad_mask=pad_mask, attn_mask=attn_mask)



# Copyright 2021 biubug6 . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.

#adapted from https://github.com/biubug6/Pytorch_Retinaface
import torch
from torch import nn
from towhee.models.retina_face.utils import conv_dw, conv_bn

class MobileNetV1(nn.Module):
    """
    MobileNetV1

    MobileNetV1 Backbone for RetinaFace.
    """
    def __init__(self):
        super().__init__()
        self.stage1 = nn.Sequential(
            conv_bn(3, 8, 2, leaky = 0.1),    # 3
            conv_dw(8, 16, 1),   # 7
            conv_dw(16, 32, 2),  # 11
            conv_dw(32, 32, 1),  # 19
            conv_dw(32, 64, 2),  # 27
            conv_dw(64, 64, 1),  # 43
        )
        self.stage2 = nn.Sequential(
            conv_dw(64, 128, 2),  # 43 + 16 = 59
            conv_dw(128, 128, 1), # 59 + 32 = 91
            conv_dw(128, 128, 1), # 91 + 32 = 123
            conv_dw(128, 128, 1), # 123 + 32 = 155
            conv_dw(128, 128, 1), # 155 + 32 = 187
            conv_dw(128, 128, 1), # 187 + 32 = 219
        )
        self.stage3 = nn.Sequential(
            conv_dw(128, 256, 2), # 219 +3 2 = 241
            conv_dw(256, 256, 1), # 241 + 64 = 301
        )
        self.avg = nn.AdaptiveAvgPool2d((1,1))
        self.fc = nn.Linear(256, 1000)

    def forward(self, x: torch.FloatTensor):
        x = self.stage1(x)
        x = self.stage2(x)
        x = self.stage3(x)
        x = self.avg(x)
        # x = self.model(x)
        x = x.view(-1, 256)
        x = self.fc(x)
        return x



# original code from https://github.com/krasserm/perceiver-io
# modified by Zilliz


from torch import nn


class Residual(nn.Module):
    """
    Residual module for Perceiver https://arxiv.org/pdf/2103.03206.pdf.

    Args:
        module (`nn.Module`):
            nn.Module.
        dropout (`nn.Module`):
            Dropout probability.
    """
    def __init__(self, module: nn.Module, dropout: float):
        super().__init__()
        self.module = module
        self.dropout = nn.Dropout(p=dropout)
        self.dropout_p = dropout

    def forward(self, *args, **kwargs):
        x = self.module(*args, **kwargs)
        return self.dropout(x) + args[0]



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import signal
import queue
import threading
import traceback
from concurrent.futures import ThreadPoolExecutor

from towhee.engine.operator_pool import OperatorPool
from towhee.engine.operator_runner.runner_base import RunnerBase
from towhee.utils.log import engine_log


class ThreadPoolTaskExecutor(threading.Thread):
    """
    A thread pool executor. Each device has one Executor.

    Args:
        name: (`str`)
            Name of the device for which the executor will run. Example device name:
                dev_name_0 = 'cpu:0'
                dev_name_1 = 'gpu:0'
        cache_path: (`str`)
            Local path for which operators are stored. Defaults to
            `$HOME/.towhee/operators`.
    """
    def __init__(self, name: str, cache_path: str = None):
        super().__init__()
        self._name = name
        self._task_queue = queue.Queue()
        self._op_pool = OperatorPool(cache_path=cache_path)
        self._thread_pool = ThreadPoolExecutor()
        self._is_run = True
        self.setDaemon(True)

    @property
    def name(self):
        return self._name

    @property
    def num_tasks(self):
        return self._task_queue.qsize()

    def push_task(self, task: RunnerBase) -> bool:
        """Push a task to the end of the task queue.

        Args:
            task:
                Pre-initialized `Task` object to push onto the queue.
        """
        return self._task_queue.put(task)

    def execute(self, runner: RunnerBase):
        try:
            op = self._op_pool.acquire_op(runner.hub_op_id, runner.op_args, runner.tag)
            runner.set_op(op)
            runner.process()
            if runner.is_end() and runner.op is not None:
                runner.unset_op()
                self._op_pool.release_op(op)
        except Exception as e:  # pylint: disable=broad-except
            engine_log.error(traceback.format_exc())
            engine_log.error(e)
            os.kill(os.getpid(), signal.SIGINT)

    def run(self):
        """
        Runs the execution loop.
        """

        while self._is_run:

            # If there are no tasks in the queue, will blocking until sth put into the queue.
            # When the executor ends, self.stop() function will put None into the queue,
            # so if get None, break the loop.
            runner = self._task_queue.get()

            if self._is_run and runner is not None:
                self._thread_pool.submit(self.execute, runner)
            else:
                break

    def stop(self):
        """
        Sets a flag, which stops the execution loop after a period of time.
        """
        self._is_run = False
        self._task_queue.put(None)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable, Generic
from typing import Any, TypeVar

A = TypeVar('A', covariant=True)
B = TypeVar('B')
T = TypeVar('T')


class _Reason:
    """
    reason for `Empty` value
    """
    def __init__(self, x: Any, e: Exception) -> None:
        self._value = x
        self._exception = e

    @property
    def value(self):
        return self._value

    @property
    def exception(self):
        return self._exception


class Option(Generic[A]):
    """
    Functional-style error handling.

    Option[A] = Some(A) or Empty[A]
    1. Some(A): just a container for result;
    2. Empty[A]: result is empty, because of input error or computation error;

    Examples:

    >>> a = Some(10)
    >>> a.map(lambda x: x/2.0)
    Some(5.0)

    >>> a.map(lambda x: x/0)
    Empty()

    >>> b = Empty()
    >>> b.map(lambda x: x/2.0)
    Empty()
    """

    @staticmethod
    def of(x: T):
        """Return a boxed value

        Args:
            x (T): input value

        Returns:
            Some(T): boxed value
        """
        return Some(x)

    @staticmethod
    def empty():
        """Return an empty value

        Returns:
            Empty: empty value
        """
        return Empty()

    def flat_map(self, f: Callable[[A], 'Option[B]']) -> 'Option[B]':
        """Apply boxed version of callable

        Args:
            f (Callable[[A], Option[B]]): boxed version of callable

        Returns:
            Option[B]: boxed value
        """
        if isinstance(self, Some):
            return f(self._value)
        else:
            return self

    def map(self, f: Callable[[A], 'B']) -> 'Option[B]':
        """Apply function to value

        Args:
            f (Callable[[A], B]): unboxed function

        Returns:
            Option[B]: boxed return value
        """
        def wrapper(x):
            try:
                return Some(f(x))
            except Exception as e: # pylint: disable=broad-except
                return Empty(x, e)

        return self.flat_map(wrapper)

    def is_empty(self):
        """Return True if the value is empty.
        """
        return isinstance(self, Empty)

    def is_some(self):
        """Return True if the value is boxed value.
        """
        return isinstance(self, Some)

    def get_or_else(self, default):
        """Return unboxed value, or default is the value is empty.
        """
        if self.is_some():
            return self.get()
        return default


class Some(Option[A]):
    """
    `Some` value for `Option`
    """
    def __init__(self, x: A) -> None:
        self._value = x

    def __repr__(self) -> str:
        return 'Some({})'.format(self._value)

    def flat_map(self, f: Callable[[A], 'Option[B]']) -> 'Option[B]':
        return f(self._value)

    def get(self):
        """Return unboxed value
        """
        return self._value


class Empty(Option[A]):
    """
    `Empty` value for `Option`
    """
    def __init__(self, x: Any = None, e: Exception = None) -> None:
        self._reason = _Reason(x, e)

    def __repr__(self) -> str:
        return 'Empty()'

    def flat_map(self, f: Callable[[A], 'Option[B]']) -> 'Option[B]':
        return self

    def get(self):
        """Return the reason of the empty value.
        """
        return self._reason


empty = Empty()

if __name__ == '__main__':
    import doctest

    doctest.testmod(verbose=False)



import torch

from torch import nn


class SPP(nn.Module):
    """
    Spatial Pyramid Pooling in Deep Convolutional Networks
    """
    def __init__(self, kernel_size1, kernel_size2, kernel_size3):
        super().__init__()
        self.pool1 = nn.MaxPool2d(kernel_size1, stride=1, padding=kernel_size1 // 2)
        self.pool2 = nn.MaxPool2d(kernel_size2, stride=1, padding=kernel_size2 // 2)
        self.pool3 = nn.MaxPool2d(kernel_size3, stride=1, padding=kernel_size3 // 2)

    def forward(self, x):
        x1 = self.pool1(x)
        x2 = self.pool2(x)
        x3 = self.pool3(x)
        return torch.cat([x, x1, x2, x3], dim=1)



# coding=utf-8
# Copyright 2020-present the HuggingFace Inc. team and 2021 Zilliz.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pandas as pd
import torch

from pandas import Series
from torch.utils.data import Dataset
from PIL import Image
from torchvision import transforms
from typing import Tuple


class PyTorchImageDataset(Dataset):
    """
        PyTorchImageDataset is a dataset class for training.

        Args:
            image_path (:obj:`str`):
                Path to the images for your dataset.

            label_file (:obj:`str`):
                 Path to your label file. The label file should be a csv file. The columns in this file should be
                 [image_name, category], 'image_name' is the path of your images, 'category' is the label of accordance
                 image. For example: [image_name, dog] for one row. Note that the first row should be[image_name, category]

            data_transform (:obj:`Compose`):
                PyTorch transform of the input images.
    """

    def __init__(self, image_path: str, label_file: str, data_transform: transforms.Compose = None):
        self.image_path = image_path
        self.label_file = label_file
        self.data_transform = data_transform

        df = pd.read_csv(self.label_file)
        image_names = Series.to_numpy(df['image_name'])
        for i in range(len(image_names)):
            if os.path.splitext(image_names[i])[1] == '':
                image_names[i] += '.jpg'
        images = image_names.tolist()
        self.images = [os.path.join(self.image_path, i) for i in images]

        categories = Series.to_numpy(df['category'])
        # Count the categories
        breed_set = set(categories)
        breed_list = list(breed_set)
        dic = {}
        for i in range(len(breed_list)):
            dic[breed_list[i]] = i
        self.labels = [dic[categories[i]] for i in range(len(categories))]
        self.num_classes = len(breed_list)

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, int]:
        label = self.labels[index]
        fn = self.images[index]
        img = Image.open(fn)
        if self.data_transform:
            img = self.data_transform(img)

        return (img, label)

    def __len__(self) -> int:
        return len(self.labels)



# Copyright 2021 biubug6 . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.

#adapted from https://github.com/biubug6/Pytorch_Retinaface
import torch
import torch.nn.functional as F
from torch import nn

from towhee.models.retina_face.utils import conv_bn, conv_bn_no_relu

class SSH(nn.Module):
    """
    SSH Module

    Single stage headless Module inspired by SSH: Single Stage Headless Face Detector.
    Described in: https://arxiv.org/abs/1708.03979.

    Args:
        in_channel (`int`):
            number of input channels.
        out_channel (`int`):
            number of output channels.
    """
    def __init__(self, in_channel: int, out_channel: int):
        super().__init__()
        assert out_channel % 4 == 0
        leaky = 0
        if out_channel <= 64:
            leaky = 0.1
        self.conv3x3 = conv_bn_no_relu(in_channel, out_channel//2, stride=1)

        self.conv5x5_1 = conv_bn(in_channel, out_channel//4, stride=1, leaky = leaky)
        self.conv5x5_2 = conv_bn_no_relu(out_channel//4, out_channel//4, stride=1)

        self.conv7x7_2 = conv_bn(out_channel//4, out_channel//4, stride=1, leaky = leaky)
        self.conv7x7_3 = conv_bn_no_relu(out_channel//4, out_channel//4, stride=1)

    def forward(self, x: torch.FloatTensor):
        conv3x3 = self.conv3x3(x)

        conv5x5_1 = self.conv5x5_1(x)
        conv5x5 = self.conv5x5_2(conv5x5_1)

        conv7x7_2 = self.conv7x7_2(conv5x5_1)
        conv7x7 = self.conv7x7_3(conv7x7_2)

        out = torch.cat([conv3x3, conv5x5, conv7x7], dim=1)
        out = F.relu(out)
        return out



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from typing import overload


class Entity:
    """
    Entity is the basic data type in DataCollection.

    Users can create an Entity with free schema, which means there is no limit on attribute name and type.
    """
    @overload
    def __init__(self):
        """
        Create an empty entity.
        """
        ...

    def __init__(self, **kwargs):
        """
        Create an Entity with given attributes.
        """

        for k, v in kwargs.items():
            self.__setattr__(k, v)

    def __repr__(self):
        """
        Define the representation of the Entity.
        """
        content = str(self.__dict__.keys())
        return f'<{self.__class__.__name__} {content.strip()}>'

    def __str__(self) -> str:
        return json.dumps({k: getattr(self, k) for k in self.__dict__})

    def to_dict(self):
        return self.__dict__



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import requests
import os
import logging
import json
from typing import Dict, Set, Any

from towhee.hparam import param_scope, HyperParameter
from towhee.utils.yaml_utils import dump_yaml, load_yaml


class BaseRepr:
    """
    Base representation from which all other representation objects inherit.

    Primarily implements automatic serialization into YAML/YAML-like string formats,
    along with defining other universally used properties.

    Args:
        name (`str`):
            Name of the internal object described by this representation.
    """
    def __init__(self, name: str):
        self._name = name

    @property
    def name(self):
        return self._name

    @staticmethod
    def is_valid(info: Dict[str, Any], essentials: Set[str]) -> bool:
        """
        Check if the src is a valid YAML file to describe a component in Towhee.

        Args:
            info (`Dict[str, Any]`):
                The dict loaded from the source file.
            essentials (`Set[str]`):
                The essential keys that a valid YAML file should contain.

        Returns:
            (`bool`)
                Return `True` if the src file is a valid YAML file to describe a
                component in Towhee, else `False`.
        """
        info_keys = set(info.keys())
        if not isinstance(info, dict) or not essentials.issubset(info_keys):
            logging.error('Info [%s] is not valid, lack attr [%s]', str(info), essentials - info_keys)
            return False
        return True

    @staticmethod
    def render_template(string: str):
        string = string.read() if hasattr(string, 'read') else string
        retval = load_yaml(string)
        retval = param_scope(**retval)
        variables = retval().variables({})
        with param_scope(**variables) as hp:
            variables.update(hp().variables({}))
            rendered = string.format(**variables)
        return rendered

    @staticmethod
    def inject_template(info: Dict[str, Any]) -> Dict:
        def inject(op, injections):
            if op['name'] in injections:
                patch = injections[op['name']]
                op = HyperParameter(**op)
                op.update(patch)
            return op

        with param_scope() as hp:
            if hp().injections(None) is not None:
                info['operators'] = [inject(op, hp().injections()) for op in info['operators']]
                if 'ir' in info:
                    del info['ir']
                info = json.loads(json.dumps(info))
                info['ir'] = dump_yaml(info)
        return info

    @staticmethod
    def load_str(string: str) -> dict:
        """
        Load the representation(s) information from a YAML file (pre-loaded as string).

        Args:
            string (`str`):
                The string pre-loaded from a YAML.

        Returns:
            (`dict`)
                The dict loaded from the YAML file that contains the representation
                information.
        """
        rendered = BaseRepr.render_template(string)
        info = load_yaml(rendered)
        info['ir'] = rendered
        info = BaseRepr.inject_template(info)
        return info

    @staticmethod
    def load_file(file: str) -> dict:
        """
        Load the representation(s) information from a local YAML file.

        Args:
            file (`str`):
                The file path.

        Returns:
            (`dict`)
                The dict loaded from the YAML file that contains the representation
                information.
        """
        with open(file, 'r', encoding='utf-8') as f:
            return BaseRepr.load_str(f)

    @staticmethod
    def load_url(url: str) -> dict:
        """
        Load the representation information from a remote YAML file.

        Args:
            url (`str`):
                The url points to the remote YAML file.

        Returns:
            (`dict`)
                The dict loaded from the YAML file that contains the representation
                information.
        """
        src = requests.get(url, timeout=5).text
        return BaseRepr.load_str(src)

    @staticmethod
    def load_src(file_or_src: str) -> dict:
        """
        Load the information for the representation. We support file from local
        file/HTTP/HDFS.

        Args:
            file_or_src (`str`):
                The source YAML file or the URL points to the source file or a str
                loaded from source file.

        returns:
            (`dict`)
                The YAML file loaded as dict.
        """
        # If `file_or_src` is a loacl file.
        if os.path.isfile(file_or_src):
            return BaseRepr.load_file(file_or_src)
        # If `file_or_src` from HTTP.
        elif file_or_src.lower().startswith('http'):
            return BaseRepr.load_url(file_or_src)
        # If `file_or_src` is neither a file nor url.
        return BaseRepr.load_str(file_or_src)



# Copyright 2021 Ross Wightman . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.

"""PyTorch implementations of DropBlock and DropPath (Stochastic Depth) regularization layers.

PyTorch implementations of DropBlock and DropPath (Stochastic Depth) regularization layers.

Papers:
DropBlock: A regularization method for convolutional networks (https://arxiv.org/abs/1810.12890)

Copyright 2020 Ross Wightman
"""
import torch
from torch import nn
import torch.nn.functional as F

def drop_block_2d(
        x, drop_prob: float = 0.1, block_size: int = 7,  gamma_scale: float = 1.0,
        with_noise: bool = False, inplace: bool = False, batchwise: bool = False):
    """ DropBlock. See https://arxiv.org/pdf/1810.12890.pdf

    DropBlock with an experimental gaussian noise option. This layer has been tested on a few training
    runs with success, but needs further validation and possibly optimization for lower runtime impact.
    """
    _, c, h, w = x.shape
    total_size = w * h
    clipped_block_size = min(block_size, min(w, h))
    # seed_drop_rate, the gamma parameter
    gamma = gamma_scale * drop_prob * total_size / clipped_block_size ** 2 / (
        (w - block_size + 1) * (h - block_size + 1))

    # Forces the block to be inside the feature map.
    w_i, h_i = torch.meshgrid(torch.arange(w).to(x.device), torch.arange(h).to(x.device))
    valid_block = ((w_i >= clipped_block_size // 2) & (w_i < w - (clipped_block_size - 1) // 2)) & \
                  ((h_i >= clipped_block_size // 2) & (h_i < h - (clipped_block_size - 1) // 2))
    valid_block = torch.reshape(valid_block, (1, 1, h, w)).to(dtype=x.dtype)

    if batchwise:
        # one mask for whole batch, quite a bit faster
        uniform_noise = torch.rand((1, c, h, w), dtype=x.dtype, device=x.device)
    else:
        uniform_noise = torch.rand_like(x)
    block_mask = ((2 - gamma - valid_block + uniform_noise) >= 1).to(dtype=x.dtype)
    block_mask = -F.max_pool2d(
        -block_mask,
        kernel_size=clipped_block_size,  # block_size,
        stride=1,
        padding=clipped_block_size // 2)

    if with_noise:
        normal_noise = torch.randn((1, c, h, w), dtype=x.dtype, device=x.device) if batchwise else torch.randn_like(x)
        if inplace:
            x.mul_(block_mask).add_(normal_noise * (1 - block_mask))
        else:
            x = x * block_mask + normal_noise * (1 - block_mask)
    else:
        normalize_scale = (block_mask.numel() / block_mask.to(dtype=torch.float32).sum().add(1e-7)).to(x.dtype)
        if inplace:
            x.mul_(block_mask * normalize_scale)
        else:
            x = x * block_mask * normalize_scale
    return x


def drop_block_fast_2d(
        x: torch.Tensor, drop_prob: float = 0.1, block_size: int = 7,
        gamma_scale: float = 1.0, with_noise: bool = False, inplace: bool = False, batchwise: bool = False):
    """ DropBlock. See https://arxiv.org/pdf/1810.12890.pdf

    DropBlock with an experimental gaussian noise option. Simplied from above without concern for valid
    block mask at edges.
    """
    _, c, h, w = x.shape
    total_size = w * h
    clipped_block_size = min(block_size, min(w, h))
    gamma = gamma_scale * drop_prob * total_size / clipped_block_size ** 2 / (
            (w - block_size + 1) * (h - block_size + 1))

    if batchwise:
        # one mask for whole batch, quite a bit faster
        block_mask = torch.rand((1, c, h, w), dtype=x.dtype, device=x.device) < gamma
    else:
        # mask per batch element
        block_mask = torch.rand_like(x) < gamma
    block_mask = F.max_pool2d(
        block_mask.to(x.dtype), kernel_size=clipped_block_size, stride=1, padding=clipped_block_size // 2)

    if with_noise:
        normal_noise = torch.randn((1, c, h, w), dtype=x.dtype, device=x.device) if batchwise else torch.randn_like(x)
        if inplace:
            x.mul_(1. - block_mask).add_(normal_noise * block_mask)
        else:
            x = x * (1. - block_mask) + normal_noise * block_mask
    else:
        block_mask = 1 - block_mask
        normalize_scale = (block_mask.numel() / block_mask.to(dtype=torch.float32).sum().add(1e-7)).to(dtype=x.dtype)
        if inplace:
            x.mul_(block_mask * normalize_scale)
        else:
            x = x * block_mask * normalize_scale
    return x


class DropBlock2d(nn.Module):
    """ DropBlock. See https://arxiv.org/pdf/1810.12890.pdf
    """
    def __init__(self,
                 drop_prob=0.1,
                 block_size=7,
                 gamma_scale=1.0,
                 with_noise=False,
                 inplace=False,
                 batchwise=False,
                 fast=True):
        super().__init__()
        self.drop_prob = drop_prob
        self.gamma_scale = gamma_scale
        self.block_size = block_size
        self.with_noise = with_noise
        self.inplace = inplace
        self.batchwise = batchwise
        self.fast = fast  # FIXME finish comparisons of fast vs not

    def forward(self, x):
        if not self.training or not self.drop_prob:
            return x
        if self.fast:
            return drop_block_fast_2d(
                x, self.drop_prob, self.block_size, self.gamma_scale, self.with_noise, self.inplace, self.batchwise)
        else:
            return drop_block_2d(
                x, self.drop_prob, self.block_size, self.gamma_scale, self.with_noise, self.inplace, self.batchwise)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path


class HubFileUtils:
    """
    Class for towhee file about hub token.
    """
    def __init__(self):
        self._file = (Path.home() / '.towhee/token').resolve()
        self._token = self.get_token()

    @property
    def token(self):
        return self._token

    @property
    def file(self):
        return self._file

    def set_token(self, token) -> None:
        self._token = token

    def get_token(self) -> str:
        try:
            return self._file.read_text()
        except FileNotFoundError:
            pass

    def save(self) -> None:
        if not self._file.parent.exists():
            self._file.parent.mkdir(parents=True)
        self._file.write_text(self._token)

    def delete(self) -> None:
        self._file.unlink()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Tuple, Dict

from towhee.dataframe import DataFrame
from towhee.utils.log import engine_log


class DataFrameWriter(ABC):
    """
    Dataframe writer base.
    """

    def __init__(self, output_df: DataFrame):
        self._output_df = output_df

    def write(self, output_data: any) -> None:
        if output_data is not None:
            self._write(output_data)
        else:
            engine_log.debug('Ignore None data which try to write to dataframe %s',
                             self._output_df.name)

    @abstractmethod
    def _write(self, output_data: any):
        raise NotImplementedError

    def close(self):
        self._output_df.seal()

class DictDataFrameWriter(DataFrameWriter):
    """
    Df writer

    Write dict to the next dataframe.
    """

    def _write(self, output_data: Dict[str, any]) -> None:
        self._output_df.put(output_data)


class RowDataFrameWriter(DataFrameWriter):
    """
    Filter dataframe writer.

    Write prev dataframe's one row to the next dataframe.
    """

    def _write(self, output_data: Tuple) -> None:
        self._output_df.put(output_data)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, Any, Optional, Set, Union, List

from towhee.functional.entity import Entity
from towhee.hparam import param_scope


def _select_callback(self):
    # pylint: disable=unused-argument
    # pylint: disable=protected-access
    def wrapper(_: str, index, *arg, **kws):
        if isinstance(index, str):
            index = (index,)
        if index is None and arg is not None and len(arg) > 0:
            index = arg

        def inner(entity: Entity):
            if index is not None:
                data = {column: getattr(entity, column) for column in index}
                return Entity(**data)
            return entity

        return self._factory(map(inner, self._iterable))
    return wrapper

class EntityMixin:
    """
    Mixin to help deal with Entity.

    Examples:

    1. define an operator with `register` decorator

    >>> from towhee import register
    >>> from towhee import DataCollection
    >>> @register
    ... def add_1(x):
    ...     return x+1

    2. apply the operator to named field of entity and save result to another named field

    >>> (
    ...     DataCollection([dict(a=1, b=2), dict(a=2, b=3)])
    ...         .as_entity()
    ...         .add_1['a', 'c']() # <-- use field `a` as input and filed `c` as output
    ...         .as_str()
    ...         .to_list()
    ... )
    ['{"a": 1, "b": 2, "c": 2}', '{"a": 2, "b": 3, "c": 3}']

    Select the entity on the specified fields.

    Examples:

    1. Select the entity on one specified field:

    >>> from towhee import Entity
    >>> from towhee import DataCollection
    >>> dc = DataCollection([Entity(a=i, b=i, c=i) for i in range(2)])
    >>> dc.select['a']().to_list()
    [<Entity dict_keys(['a'])>, <Entity dict_keys(['a'])>]

    2. Select multiple fields and unpack the entity:

    >>> (
    ...     DataCollection([Entity(a=i, b=i, c=i) for i in range(5)])
    ...         .select['a', 'b']()
    ...         .as_raw()
    ...         .to_list()
    ... )
    [(0, 0), (1, 1), (2, 2), (3, 3), (4, 4)]

    3. Another field selection syntax (not suggested):

    >>> (
    ...     DataCollection([Entity(a=i, b=i, c=i) for i in range(5)])
    ...         .select('a', 'b')
    ...         .as_raw()
    ...         .to_list()
    ... )
    [(0, 0), (1, 1), (2, 2), (3, 3), (4, 4)]
    """
    def __init__(self):
        super().__init__()
        self.select = param_scope().callholder(_select_callback(self))

    # pylint: disable=invalid-name
    def fill_entity(self, _DefaultKVs: Optional[Dict[str, Any]] = None, _ReplaceNoneValue: bool = False, **kws):
        """
        When DataCollection's iterable exists of Entities and some indexes missing, fill default value for those indexes.

        Args:
            _ReplaceNoneValue (`bool`):
                Whether to replace None in Entity's value.
            _DefaultKVs (`Dict[str, Any]`):
                The key-value pairs stored in a dict.
        """
        if _DefaultKVs:
            kws.update(_DefaultKVs)

        def fill(entity: Entity):
            for k, v in kws.items():
                if not hasattr(entity, k):
                    setattr(entity, k, v)
                if _ReplaceNoneValue and v is None:
                    setattr(entity, k, 0)
            return entity

        return self._factory(map(fill, self._iterable))

    def as_entity(self, schema: Optional[List[str]]=None):
        """
        Convert elements into Entities.

        Args:
            schema (Optional[List[str]]):
                schema contains field names.

        Examples:
        1. convert dicts into entities:

        >>> from towhee import DataCollection
        >>> (
        ...     DataCollection([dict(a=1, b=2), dict(a=2, b=3)])
        ...         .as_entity()
        ...         .as_str()
        ...         .to_list()
        ... )
        ['{"a": 1, "b": 2}', '{"a": 2, "b": 3}']

        2. convert tuples into entities:

        >>> from towhee import DataCollection
        >>> (
        ...     DataCollection([(1, 2), (2, 3)])
        ...         .as_entity(schema=['a', 'b'])
        ...         .as_str()
        ...         .to_list()
        ... )
        ['{"a": 1, "b": 2}', '{"a": 2, "b": 3}']

        3. convert single value into entities:

        >>> from towhee import DataCollection
        >>> (
        ...     DataCollection([1, 2])
        ...         .as_entity(schema=['a'])
        ...         .as_str()
        ...         .to_list()
        ... )
        ['{"a": 1}', '{"a": 2}']
        """

        if schema is None:
            def inner(x):
                return Entity(**x)
        else:
            def inner(x):
                if len(schema) == 1:
                    x = (x, )
                data = dict(zip(schema, x))
                return Entity(**data)
        return self._factory(map(inner, self._iterable))

    def as_raw(self):
        """
        Convert entitis into raw python values

        Examples:

        1. unpack multiple values from entities:

        >>> from towhee import DataCollection
        >>> (
        ...     DataCollection([(1, 2), (2, 3)])
        ...         .as_entity(schema=['a', 'b'])
        ...         .as_raw()
        ...         .to_list()
        ... )
        [(1, 2), (2, 3)]

        2. unpack single value from entities:

        >>> (
        ...     DataCollection([1, 2])
        ...         .as_entity(schema=['a'])
        ...         .as_raw()
        ...         .to_list()
        ... )
        [1, 2]
        """

        def inner(x):
            if len(x.__dict__) == 1:
                return list(x.__dict__.values())[0]
            return tuple(getattr(x, name) for name in x.__dict__)
        return self._factory(map(inner, self._iterable))

    def replace(self, **kws):
        """
        Replace specific attributes with given vlues.
        """
        def inner(entity: Entity):
            for index, convert_dict in kws.items():
                origin_value = getattr(entity, index)
                if origin_value in convert_dict:
                    setattr(entity, index, convert_dict[origin_value])

            return entity

        return self._factory(map(inner, self._iterable))

    def dropna(self, na: Set[str] = {'', None}) -> Union[bool, 'DataCollection']:  # pylint: disable=dangerous-default-value
        """
        Drop entities that contain some specific values.

        Args:
            na (`Set[str]`):
                Those entities contain values in na will be dropped.
        """
        def inner(entity: Entity):
            for val in entity.__dict__.values():
                if val in na:
                    return False

            return True

        return self._factory(filter(inner, self._iterable))
    def rename(self, column: Dict[str, str]):
        """
        Rename an column in DataCollection.

        Args:
            column (`Dict[str, str]`):
                The columns to rename and their corresponding new name.
        """
        def inner(x):
            for key in column:
                x.__dict__[column[key]] = x.__dict__.pop(key)
            return x

        return self._factory(map(inner, self._iterable))


if __name__ == '__main__':
    import doctest

    doctest.testmod(verbose=False)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import shutil
import argparse
from pathlib import Path
from requests.exceptions import HTTPError
from urllib.parse import urlsplit

from towhee.hub.operator_manager import OperatorManager
from towhee.hub.pipeline_manager import PipelineManager
from towhee.utils.hub_utils import HubUtils
from towhee.utils.hub_file_utils import HubFileUtils
from towhee.utils.repo_normalize import RepoNormalize
from towhee.utils.git_utils import GitUtils

parser = argparse.ArgumentParser(add_help=False)
parser.add_argument('uri', help='Repo uri, such as author/repo-name or repo-name(author defaults to login account)')
parser.add_argument('-d', '--dir', default='.', help='optional, directory to the Repo file, defaults to \'.\'')
parser.add_argument('--local', action='store_true', help='optional, create and init repo in local')
parser.add_argument('--plain', action='store_true', help='optional, just create repo with init file')


class RepoCommand:
    """
    Implementation for subcmd `towhee develop` and `towhee install`.
    Setup repo to `towheeoperator.{self._args.namespace}_{self._args.repo_name}' package with pypi methods.
    """

    def __init__(self, args) -> None:
        self._args = args
        self.file = HubFileUtils()
        self.hub = HubUtils()

    def __call__(self) -> None:
        self.uri = RepoNormalize(self._args.uri).parse_uri()
        self.token = self.file.get_token()
        if not self._args.local:
            if self.token is None:
                print('You have not log in, please run `towhee login` first.')
                sys.exit()
            else:
                author = self.hub.get_user(self.token).json()['username']
                if '/' not in urlsplit(self._args.uri):
                    self._args.uri = f'{author}/{self._args.uri}'
                self.uri = RepoNormalize(self._args.uri).parse_uri()
                print(f'creating repo with username: {self.uri.author}, and repo name: {self.uri.repo}\n')
                if author != self.uri.author:
                    print(f'Authenticated user: {author} does not match the username: {self.uri.author}, please specify the correct author.')
                    sys.exit()
        try:
            if self._args.action == 'create-op':
                self.create_op()
            if self._args.action == 'create-pipeline':
                self.create_pipeline()
            elif self._args.action == 'generate-yaml':
                print('Generating yaml for repo...')
                self.generate_yaml()
                print('Done')
        except HTTPError as e:
            print(e)

    @staticmethod
    def install(subparsers):
        op_parser = argparse.ArgumentParser(add_help=False)
        op_parser.add_argument('-t', '--type', choices=['pyop', 'nnop'], default='nnop',
                               help='optional, operator repo type in [\'pyop\', \'nnop\'] for init file, defaults to \'nnop\'')
        op_parser.add_argument('-f', '--framework', default='pytorch', help='optional, framework of nnoperator, defaults to \'pytorch\'')

        subparsers.add_parser('create-op', parents=[op_parser, parser], help='hub-repo command: create operator and generate init file')

    def create_op(self):
        op_manager = OperatorManager(self.uri.author, self.uri.repo)
        repo_path = Path(self._args.dir) / self.uri.repo
        if not self._args.local:
            self.create_repo(op_manager, repo_path)
        else:
            repo_path.mkdir(parents=True)
        if self._args.local or not self._args.plain:
            self.init_repo(op_manager, repo_path)

    def create_pipeline(self):
        pipeline_manager = PipelineManager(self.uri.author, self.uri.repo)
        repo_path = Path(self._args.dir) / self.uri.repo
        if not self._args.local:
            self.create_repo(pipeline_manager, repo_path)
        else:
            repo_path.mkdir(parents=True)
        if self._args.local or not self._args.plain:
            self.init_repo(pipeline_manager, repo_path, True)

    def create_repo(self, manager, repo_path):
        link = self.uri.full_uri.split('?')[0]
        if manager.exists():
            print(f'Noting {self.uri.author}/{self.uri.repo} repo already exists: {link}\n')
        else:
            manager.create_with_token(self.token)
            print(f'Successfully create Operator in hub: {link}\n')
        if not self._args.plain:
            GitUtils(self.uri.author, self.uri.repo).clone(local_repo_path=repo_path)

    def init_repo(self, manager, repo_path, is_pipeline=False):
        print('\nInitializing the repo file structure...\n')
        if is_pipeline:
            temp_path = Path(self._args.dir) / 'pipeline_template'
            GitUtils('towhee', 'pipeline-template').clone(local_repo_path=temp_path)
            manager.init_pipeline(temp_path, repo_path)
        elif self._args.type == 'pyop':
            temp_path = Path(self._args.dir) / 'pyoperator_template'
            GitUtils('towhee', 'pyoperator-template').clone(local_repo_path=temp_path)
            manager.init_pyoperator(temp_path, repo_path)
        elif self._args.type == 'nnop':
            temp_path = Path(self._args.dir) / 'nnoperator_template'
            GitUtils('towhee', 'nnoperator-template').clone(local_repo_path=temp_path)
            manager.init_nnoperator(temp_path, repo_path, self._args.framework)
        shutil.rmtree(str(temp_path))

    def generate_yaml(self):
        repo_manager = OperatorManager(self.uri.author, self.uri.repo)
        repo_manager.generate_yaml(Path(self._args.dir))


class PipeCommand:
    """
    Implementation for subcmd `towhee develop` and `towhee install`.
    Setup repo to `towheeoperator.{self._args.namespace}_{self._args.repo_name}' package with pypi methods.
    """

    def __init__(self, args) -> None:
        self._args = args

    def __call__(self) -> None:
        RepoCommand(self._args)()

    @staticmethod
    def install(subparsers):
        subparsers.add_parser('create-pipeline', parents=[parser], help='hub-repo command: create pipeline and generate init file')



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import concurrent.futures
from queue import Queue
import asyncio
import threading

from towhee.utils.log import engine_log
from towhee.functional.option import Option, Empty, _Reason
from towhee.hparam.hyperparameter import param_scope


def _map_task(x, unary_op):
    def map_wrapper():
        try:
            if isinstance(x, Option):
                return x.map(unary_op)
            else:
                return unary_op(x)
        except Exception as e:  # pylint: disable=broad-except
            engine_log.warning(f'{e}, please check {x} with op {unary_op}. Continue...')  # pylint: disable=logging-fstring-interpolation
            return Empty(_Reason(x, e))

    return map_wrapper


class ParallelMixin:
    """
    Mixin for parallel execution.

    Examples:

    >>> from towhee import DataCollection
    >>> def add_1(x):
    ...     return x+1
    >>> result = DataCollection.range(1000).set_parallel(2).map(add_1).to_list()
    >>> len(result)
    1000

    >>> result = DataCollection.range(1000).pmap(add_1, 10).pmap(add_1, 10).to_list()
    >>> result[990:]
    [992, 993, 994, 995, 996, 997, 998, 999, 1000, 1001]
    """

    def __init__(self) -> None:
        super().__init__()
        with param_scope() as hp:
            parent = hp().data_collection.parent(None)
        if parent is not None and hasattr(parent, '_executor') and isinstance(parent._executor, concurrent.futures.ThreadPoolExecutor):
            self._backend = parent._backend
            self._executor = parent._executor
            self._num_worker = parent._num_worker

    def get_executor(self):
        if hasattr(self, '_executor') and isinstance(self._executor, concurrent.futures.ThreadPoolExecutor) :
            return self._executor
        return None

    def get_backend(self):
        if hasattr(self, '_backend')  and isinstance(self._backend, str):
            return self._backend
        return None

    def get_num_worker(self):
        if hasattr(self, '_num_worker')  and isinstance(self._num_worker, int):
            return self._num_worker
        return None

    def set_parallel(self, num_worker = 2, backend = 'thread'):
        """
        Set parallel execution for following calls.

        Examples:

        >>> from towhee import DataCollection
        >>> import threading
        >>> stage_1_thread_set = set()
        >>> stage_2_thread_set = set()
        >>> result = (
        ...     DataCollection.range(1000).stream().set_parallel(4)
        ...     .map(lambda x: stage_1_thread_set.add(threading.current_thread().ident))
        ...     .map(lambda x: stage_2_thread_set.add(threading.current_thread().ident)).to_list()
        ... )

        >>> len(stage_2_thread_set)>1
        True
        """
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=num_worker)
        self._backend = backend
        self._num_worker = num_worker
        return self

    def unset_parallel(self):
        """
        Unset parallel execution for following calls.

        Examples:

        >>> from towhee import DataCollection
        >>> import threading
        >>> stage_1_thread_set = {threading.current_thread().ident}
        >>> stage_2_thread_set = {threading.current_thread().ident}
        >>> result = (
        ...     DataCollection.range(1000).stream().set_parallel(4)
        ...     .map(lambda x: stage_1_thread_set.add(threading.current_thread().ident))
        ...     .unset_parallel()
        ...     .map(lambda x: stage_2_thread_set.add(threading.current_thread().ident)).to_list()
        ... )

        >>> len(stage_1_thread_set)>1
        True
        >>> len(stage_2_thread_set)>1
        False
        """
        self._backend = None
        self._num_wokrker = None
        self._executor = None
        return self

    def split(self, count):
        """
        Split a dataframe into multiple dataframes.

        Args:
            count (int): how many resulting DCs;

        Returns:
            [DataCollection, ...]: copies of DC;

        Examples:

        1. Split:
        >>> from towhee import DataCollection
        >>> dc = DataCollection([0, 1, 2, 3, 4]).stream()
        >>> a, b, c = dc.split(3)
        >>> a.zip(b, c).to_list()
        [(0, 0, 0), (1, 1, 1), (2, 2, 2), (3, 3, 3), (4, 4, 4)]
        """
        queues = [Queue(maxsize=1) for _ in range(count)]
        loop = asyncio.new_event_loop()

        def inner(queue):
            while True:
                x = queue.get()
                if isinstance(x, EOS):
                    break
                else:
                    yield x

        async def worker():
            for x in self:
                for queue in queues:
                    queue.put(x)
            for queue in queues:
                poison = EOS()
                queue.put(poison)

        def worker_wrapper():
            loop.run_until_complete(worker())
            loop.close()

        t = threading.Thread(target=worker_wrapper)
        t.start()
        retval = [inner(queue) for queue in queues]
        return [self._factory(x) for x in retval]

    def pmap(self, unary_op, num_worker = None, backend = None):
        """
        Apply `unary_op` with parallel execution.
        Currently supports two backends, `ray` and `thread`.

        Args:
            unary_op (func): the op to be mapped;
            num_worker (int): how many threads to reserve for this op;
            backend (str): whether to use `ray` or `thread`

        Examples:

        >>> from towhee import DataCollection
        >>> import threading
        >>> stage_1_thread_set = {threading.current_thread().ident}
        >>> stage_2_thread_set = {threading.current_thread().ident}
        >>> result = (
        ...     DataCollection.range(1000).stream()
        ...     .pmap(lambda x: stage_1_thread_set.add(threading.current_thread().ident), 5)
        ...     .pmap(lambda x: stage_2_thread_set.add(threading.current_thread().ident), 4).to_list()
        ... )
        >>> len(stage_1_thread_set) > 1
        True
        >>> len(stage_2_thread_set) > 1
        True
        """
        if backend is None:
            if self.get_backend() == 'ray':
                return self._ray_pmap(unary_op, num_worker)
            else:
                return self._thread_pmap(unary_op, num_worker)
        elif backend == 'thread':
            return self._thread_pmap(unary_op, num_worker)
        elif backend == 'ray':
            return self._ray_pmap(unary_op, num_worker)

    def _thread_pmap(self, unary_op, num_worker=None):
        if num_worker is not None:
            executor = concurrent.futures.ThreadPoolExecutor(num_worker)
        elif self.get_executor() is not None:
            executor = self._executor
            num_worker = self._num_worker
        else:
            executor = concurrent.futures.ThreadPoolExecutor(2)
            num_worker = 2

        queue = Queue(num_worker)
        loop = asyncio.new_event_loop()

        def inner():
            while True:
                x = queue.get()
                if isinstance(x, EOS):
                    break
                else:
                    yield x

        async def worker():
            buff = []
            for x in self:
                if len(buff) == num_worker:
                    queue.put(await buff.pop(0))
                buff.append(loop.run_in_executor(executor, _map_task(x, unary_op)))
            while len(buff) > 0:
                queue.put(await buff.pop(0))
            queue.put(EOS())

        def worker_wrapper():
            loop.run_until_complete(worker())
            loop.close()

        t = threading.Thread(target=worker_wrapper)
        t.start()

        return self._factory(inner())

    def mmap(self, ops: list, num_worker = None, backend = None):
        """
        Apply multiple unary_op to data collection.
        Currently supports two backends, `ray` and `thread`.

        Args:
            unary_op (func): the op to be mapped;
            num_worker (int): how many threads to reserve for this op;
            backend (str): whether to use `ray` or `thread`

        Examples:

        1. Using mmap:

        >>> from towhee import DataCollection
        >>> dc = DataCollection([0,1,2,'3',4]).stream()
        >>> a, b = dc.mmap([lambda x: x+1, lambda x: x*2])
        >>> c = a.map(lambda x: x+1)
        >>> c.zip(b).to_list()
        [(2, 0), (3, 2), (4, 4), (Empty(), '33'), (6, 8)]

        2. Using map instead of mmap:

        >>> from towhee import DataCollection
        >>> dc = DataCollection.range(5).stream()
        >>> a, b, c = dc.map(lambda x: x+1, lambda x: x*2, lambda x: int(x/2))
        >>> d = a.map(lambda x: x+1)
        >>> d.zip(b, c).to_list()
        [(2, 0, 0), (3, 2, 0), (4, 4, 1), (5, 6, 1), (6, 8, 2)]

        3. DAG execution:

        >>> dc = DataCollection.range(5).stream()
        >>> a, b, c = dc.map(lambda x: x+1, lambda x: x*2, lambda x: int(x/2))
        >>> d = a.map(lambda x: x+1)
        >>> d.zip(b, c).map(lambda x: x[0]+x[1]+x[2]).to_list()
        [2, 5, 9, 12, 16]
        """
        if len(ops) == 1:
            return self._pmap(unary_op=ops[0], num_worker=num_worker, backend=backend)

        next_vals = []
        next_vals = self.split(len(ops))

        ret = []
        for i, x in enumerate(ops):
            ret.append(next_vals[i].pmap(x, num_worker=num_worker, backend=backend))
        return ret

class EOS():
    '''
    Internal object used to signify end of processing queue.
    '''
    pass


if __name__ == '__main__':  # pylint: disable=inconsistent-quotes
    import doctest
    doctest.testmod(verbose=False)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from towhee.operator import Operator, SharedType


class ConcatOperator(Operator):
    """
    Concat operator.

    Combine multiple dataframes into one. It's an internal operator, and the
    framework will set readers and writer to it, so the operator can use readers
    and wirter directly.

    Row based concat:

    Col based concat:
    """

    def __init__(self, concat_type: str = 'row') -> None:
        if concat_type.lower() == 'row':
            self._concat_func = self._row_based_concat
        elif concat_type.lower() == 'col':
            self._concat_func = self._col_based_concat
        else:
            raise RuntimeError('Unkown concat type: %s ' % concat_type)

    @staticmethod
    def _read(reader):
        try:
            data, _ = reader.read()
            return False, data
        except StopIteration:
            return True, None
        pass

    def _row_based_concat(self):
        while True:
            result = {}
            for reader in self._readers:
                is_end, data = ConcatOperator._read(reader)
                if not is_end:
                    result.update(data)
            if len(result) == 0:
                break
            self._writer.write(result)

    def _col_based_concat(self):
        raise NotImplementedError

    def __call__(self) -> bool:
        self._concat_func()

    @property
    def shared_type(self):
        return SharedType.NotReusable



import threading
from typing import List, Tuple
import weakref

from towhee.dataframe.dataframe import DataFrame, Responses

class DataFrameIterator:
    """
    Base iterator implementation. All iterators should be subclasses of `DataFrameIterator`.

    Args:
        df (`Dataframe`):
            The dataframe to iterate over.
        block (`bool`):
            Whether to block on call.
    """

    def __init__(self, df: DataFrame, block: bool = True):
        self._df_ref = weakref.ref(df)
        self._df_name = df.name
        self._offset = 0
        self._block = block
        self._id = df.register_iter()
        self._done = False
        self._event = threading.Event()

    def __iter__(self):
        return self

    def __next__(self) -> List[Tuple]:
        # The base iterator is purposely defined to have exatly 0 elements.
        raise StopIteration

    @property
    def df_name(self) -> str:
        """
        Returns the df name.
        """
        return self._df_name

    @property
    def current_index(self) -> int:
        """
        Returns the current index.
        """
        return self._offset

    @property
    def accessible_size(self) -> int:
        """
        Returns current accessible data size.
        """
        return self._df_ref().size - self._offset

    @property
    def id(self) -> int:
        return self._id

    def notify(self):
        """
        Remove unblock and remove iterator.
        """
        if self._done is not True:
            df = self._df_ref()
            df.remove_iter(self._id)
            self._done = True

class MapIterator(DataFrameIterator):
    """
    A row-based map `DataFrame` iterator.

    Args:
        df (`towhee.Dataframe`):
            The dataframe that is being iterated.
        block (`bool`):
            Whether to block when data not present.
    """
    def __init__(self, df: DataFrame, block: bool = True):
        super().__init__(df, block = block)
        self._batch_size = 1
        self._step = 1

    def __iter__(self):
        return self

    def __next__(self):
        """
        Returns:
            (`list(tuple([Any, ...])`)
                In the normal case, the iterator will return a list of tuples at each call.
            (`None`)
                In the case that the `DataFrame` is not sealed and the new rows are
                not ready yet, the iterator will return `None`.
        Raises:
            (`StopIteration`)
                The iteration end if the `DataFrame` is sealed and the last row is
                reached.
        """
        df = self._df_ref()

        if self._done:
            raise StopIteration

        code, row, _ = df.get(self._offset, count = self._batch_size, iter_id = self._id)

        if code == Responses.INDEX_GC:
            raise IndexError

        elif code == Responses.INDEX_OOB_UNSEALED:
            if self._block:
                df.notify_map_block(self._event, self._offset, self._batch_size, self._id)
                self._event.wait()
                self._event.clear()
                return self.__next__()

            return None

        elif code == Responses.APPROVED_CONTINUE:
            df.ack(self._offset + self._step, self._id)

            self._offset += self._step
            return row

        elif code == Responses.INDEX_OOB_SEALED:
            df.remove_iter(self._id)
            self._offset = 0
            self._done = True
            raise StopIteration

        elif code == Responses.KILLED:
            raise StopIteration

        else: # 'unkown_error'
            raise Exception


class BatchIterator(MapIterator):
    """
    A row-based batched map `DataFrame` iterator.

    Args:
        df (`towhee.Dataframe`):
            The dataframe that is being iterated.
        batch_size (`int`):
            How many values to read.
        step (`int`):
            How many steps to take after each read.
        block (`bool`):
            Whether to block when data not present.
    """
    def __init__(self, df: DataFrame, batch_size: int = 1, step: int = 1, block: bool = True):
        super().__init__(df, block = block)
        self._batch_size = batch_size
        self._step = step


class WindowIterator(DataFrameIterator):
    """
    A row-based window `DataFrame` iterator.

    Args:
        df (`towhee.Dataframe`):
            The dataframe that is being iterated.
        start (`int`):
            Where to start the window from.
        window_size (`int`):
            How large of a window.
        step (`int`):
            How far to iterate window per read.
        use_timestamp (`bool`):
            Whether to use timestamp instead of row_id.
        block (`bool`):
            Whether to block when data not present.
    """
    def __init__(self, df: DataFrame, start: int = 0, window_size: int = 1, step: int = None, use_timestamp: bool = False, block: bool = True):
        super().__init__(df, block = block)

        if use_timestamp:
            start *= 1000
            window_size *= 1000
            if step is not None:
                step *= 1000

        self._window_size = window_size
        self._current_window = (start, (start + window_size))
        self._step = step

        if step is None:
            self._step = window_size
        if use_timestamp:
            self._comparator = 'timestamp'
        else:
            self._comparator = 'row_id'


    def __next__(self):
        """
        Returns:
            (`list(Tuple[Any, ...])`)
                In the normal case, the iterator will return a list of `tuple` at each call.
            (`None`)
                In the case that the `DataFrame` is not sealed and the new rows are
                not ready yet, the iterator will return `None`. The caller should
                determine whether to block the iteration or exit the loop.
        Raises:
            (`StopIteration`)
                The iteration end iff the `DataFrame` is sealed and the last row is
                reached.
        """
        df = self._df_ref()

        if self._done:
            df.remove_iter(self._id)
            raise StopIteration

        code, rows, offset = df.get_window(self._current_window[0], self._current_window[1], self._step, self._comparator, self._id)
        if code in (Responses.EMPTY_SEALED, Responses.FUTURE_WINDOW_SEALED):
            self._done = True
            df.remove_iter(self._id)
            raise StopIteration

        elif code == Responses.EMPTY:
            df.notify_window_block(self._event, 'start', (self._comparator, self._current_window[0]), self._id)
            self._event.wait()
            self._event.clear()
            return self.__next__()

        elif code == Responses.WINDOW_NOT_DONE:
            df.notify_window_block(self._event, 'end', (self._comparator, self._current_window[1]), self._id)
            self._event.wait()
            self._event.clear()
            return self.__next__()

        elif code == Responses.FUTURE_WINDOW:
            df.ack(offset, self._id) #TODO: Garbage collection logic if waiting on window that is far.
            df.notify_window_block(self._event, 'start', (self._comparator, self._current_window[0]), self._id)
            self._event.wait()
            self._event.clear()
            return self.__next__()

        elif code == Responses.OLD_WINDOW:
            self._current_window = offset
            return self.__next__()

        elif code == Responses.APPROVED_DONE:
            self._done = True
            return rows

        elif code == Responses.APPROVED_CONTINUE:
            self._current_window = (self._current_window[0] + self._step, self._current_window[1] + self._step)
            df.ack(offset, self._id)
            return rows

        elif code == Responses.KILLED:
            self._done = True
            raise StopIteration

        else: # 'unkown_error'
            raise Exception



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest
from pathlib import Path
from typing import List

from setuptools import find_packages, setup
from setuptools.command.install import install


def test_suite():
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover('towhee/tests', pattern='test_*.py')
    return test_suite


def parse_requirements(file_name: str) -> List[str]:
    with open(file_name) as f:
        return [
            require.strip() for require in f
            if require.strip() and not require.startswith('#')
        ]


kws = {}
try:
    from sphinx.setup_command import \
        BuildDoc  # pylint: disable=import-outside-toplevel
    cmdclass = {'build_sphinx': BuildDoc}
    command_options = {
        'build_sphinx': {
            'project': ('setup.py', 'towhee'),
            'version': ('setup.py', '0.6.0'),
            'release': ('setup.py', '0.6.0'),
            'source_dir': ('setup.py', 'towhee/doc/source')
        }
    }
    kws = {'cmdclass': cmdclass, 'command_options': command_options}

except:  # pylint: disable=bare-except
    pass

if '--models' in sys.argv:
    sys.argv.remove('--models')
    setup(name='towhee.models',
          version='0.6.0',
          description='',
          author='Towhee Team',
          author_email='towhee-team@zilliz.com',
          use_scm_version={'local_scheme': 'no-local-version'},
          setup_requires=['setuptools_scm'],
          url='https://github.com/towhee-io/towhee',
          test_suite='setup.test_suite',
          install_requires=parse_requirements('requirements.txt'),
          extras_require={':python_version<\'3.7\'': 'importlib-resources'},
          tests_require=parse_requirements('test_requirements.txt'),
          packages=find_packages(include=['towhee.models*']),#['towhee.models'],
          package_data={'towhee.tests.test_util': ['*.yaml']},
          license='http://www.apache.org/licenses/LICENSE-2.0',
          entry_points={
              'console_scripts': ['towhee=towhee.command.cmdline:main'],
          },
          **kws)
else:
    setup(name='mytowhee',
          version='0.6.0',
          description='',
          author='Towhee Team',
          author_email='towhee-team@zilliz.com',
          use_scm_version={'local_scheme': 'no-local-version'},
          setup_requires=['setuptools_scm'],
          url='https://github.com/towhee-io/towhee',
          test_suite='setup.test_suite',
          install_requires=parse_requirements('requirements.txt'),
          extras_require={':python_version<\'3.7\'': 'importlib-resources'},
          tests_require=parse_requirements('test_requirements.txt'),
          packages=find_packages(exclude=['*test*', 'towhee.models*']),
          package_data={'towhee.tests.test_util': ['*.yaml']},
          license='http://www.apache.org/licenses/LICENSE-2.0',
          entry_points={
              'console_scripts': ['towhee=towhee.command.cmdline:main'],
          },
          **kws)



# Original pytorch implementation by:
# 'An Image Is Worth 16 x 16 Words: Transformers for Image Recognition at Scale'
#       - https://arxiv.org/abs/2010.11929
# 'How to train your ViT? Data, Augmentation, and Regularization in Vision Transformers'
#       - https://arxiv.org/abs/2106.10270
#
# Original code by / Copyright 2020, Ross Wightman.
# Modifications & additions by / Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from torch import nn
from towhee.models.layers.multi_head_attention import MultiHeadAttention
from towhee.models.layers.droppath import DropPath
from towhee.models.layers.mlp import Mlp

class Block(nn.Module):
    """
    The Transformer block.
    Args:
        dim (int): number of features
        num_heads (int): number of heads
        mlp_ratio (int): ratio of mlp layer
        qkv_bias (bool): if add bias to qkv layer
        qk_scale (float): number to scale qk
        drop_ratio (float): drop rate at the end of the block (mlp layer)
        attn_drop_ratio (float): drop rate of attention layer
        drop_path_ratio (float): drop rate of drop_path layer
        act_layer: activation layer
        norm_layer: normalization layer
    """
    def __init__(self,
                 dim,
                 num_heads,
                 mlp_ratio=4,
                 qkv_bias=False,
                 qk_scale=None,
                 drop_ratio=0,
                 attn_drop_ratio=0,
                 drop_path_ratio=0,
                 act_layer=nn.GELU,
                 norm_layer=nn.LayerNorm):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = MultiHeadAttention(
            dim=dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            attn_drop_ratio=attn_drop_ratio,
            proj_drop_ratio=drop_ratio
        )
        self.drop_path = DropPath(drop_prob=drop_path_ratio) if drop_path_ratio > 0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        self.mlp = Mlp(in_features=dim, hidden_features=int(dim * mlp_ratio), act_layer=act_layer, drop=drop_ratio)

    def forward(self, x):
        x = x + self.drop_path(self.attn(self.norm1(x)))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x

# if __name__=='__main__':
#     import torch
#     x = torch.rand(1, 197, 768)
#     model = Block(dim=768, num_heads=8)
#     out = model.forward(x)
#     print(out.shape)



# Copyright 2021 Ross Wightman . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.
from typing import Union, Callable, Type

import torch
from torch import nn
from torch.nn import functional as F

from towhee.models.layers.activations import (GELU,
                                              HardSigmoid,
                                              HardSwish,
                                              HardMish,
                                              Mish,
                                              PReLU,
                                              Sigmoid,
                                              Swish,
                                              Tanh)

from towhee.models.layers.activations import (gelu,
                                              hard_sigmoid,
                                              hard_swish,
                                              hard_mish,
                                              mish,
                                              sigmoid,
                                              swish,
                                              tanh)


# PyTorch has an optimized, native 'silu' (aka 'swish') operator as of PyTorch 1.7.
# Also hardsigmoid, hardswish, and soon mish. This code will use native version if present.
# Eventually, the custom SiLU, Mish, Hard*, layers will be removed and only native variants will be used.
_has_silu = 'silu' in dir(torch.nn.functional)
_has_hardswish = 'hardswish' in dir(torch.nn.functional)
_has_hardsigmoid = 'hardsigmoid' in dir(torch.nn.functional)
_has_mish = 'mish' in dir(torch.nn.functional)

_ACT_FN_DEFAULT = dict(
    silu=F.silu if _has_silu else swish,
    swish=F.silu if _has_silu else swish,
    mish=F.mish if _has_mish else mish,
    relu=F.relu,
    relu6=F.relu6,
    leaky_relu=F.leaky_relu,
    elu=F.elu,
    celu=F.celu,
    selu=F.selu,
    gelu=gelu,
    sigmoid=sigmoid,
    tanh=tanh,
    hard_sigmoid=F.hardsigmoid if _has_hardsigmoid else hard_sigmoid,
    hard_swish=F.hardswish if _has_hardswish else hard_swish,
    hard_mish=hard_mish
)

_ACT_FNS = (_ACT_FN_DEFAULT,)

_ACT_LAYER_DEFAULT = dict(
    silu=nn.SiLU if _has_silu else Swish,
    swish=nn.SiLU if _has_silu else Swish,
    mish=nn.Mish if _has_mish else Mish,
    relu=nn.ReLU,
    relu6=nn.ReLU6,
    leaky_relu=nn.LeakyReLU,
    elu=nn.ELU,
    prelu=PReLU,
    celu=nn.CELU,
    selu=nn.SELU,
    gelu=GELU,
    sigmoid=Sigmoid,
    tanh=Tanh,
    hard_sigmoid=nn.Hardsigmoid if _has_hardsigmoid else HardSigmoid,
    hard_swish=nn.Hardswish if _has_hardswish else HardSwish,
    hard_mish=HardMish,
)

_ACT_LAYERS = (_ACT_LAYER_DEFAULT,)

for a in _ACT_FNS:
    a.setdefault('hardsigmoid', a.get('hard_sigmoid'))
    a.setdefault('hardswish', a.get('hard_swish'))

def get_act_fn(name: Union[Callable, str] = 'relu'):
    """
    Activation Function Factory

    Fetching activation fns by name with this function allows export or torch script friendly
    functions to be returned dynamically based on current config.
    """
    if not name:
        return None
    if callable(name) is True:
        return name
    return _ACT_FN_DEFAULT[name]

def get_act_layer(name: Union[Type[nn.Module], str] = 'relu'):
    """
    Activation Layer Factory
    Fetching activation layers by name with this function allows export or torch script friendly
    functions to be returned dynamically based on current config.
    """
    if not name:
        return None
    if isinstance(name, type):
        return name
    return _ACT_LAYER_DEFAULT[name]

def create_act_layer(name: Union[nn.Module, str], inplace=None, **kwargs):
    act_layer = get_act_layer(name)
    if act_layer is None:
        return None
    return act_layer(**kwargs) if inplace is None else act_layer(inplace=inplace, **kwargs)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable

from towhee.engine import register

# pylint: disable=import-outside-toplevel
# pylint: disable=invalid-name

@register(name='builtin/runas_op')
class runas_op:
    """
    Convert a user-defined function as an operator and execute.

    Args:
        func (`Callable`):
            The user-defined function.

    Examples:

    >>> from towhee.functional import DataCollection
    >>> from towhee.functional.entity import Entity
    >>> entities = [Entity(a=i, b=i) for i in range(5)]
    >>> dc = DataCollection(entities)
    >>> res = dc.runas_op['a', 'b'](func=lambda x: x - 1).to_list()
    >>> res[0].a == res[0].b + 1
    True
    """
    def __init__(self, func: Callable):
        self._func = func

    def __call__(self, *args, **kws):
        return self._func(*args, **kws)


# __test__ = {'run': run.__doc__}

if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=False)



import importlib

import importlib_metadata
from towhee.utils.log import trainer_log

_captum_available = importlib.util.find_spec("captum") is not None
try:
    _captum_version = importlib_metadata.version("captum")
    trainer_log.info("Successfully imported captum version %s", _captum_version)
except importlib_metadata.PackageNotFoundError:
    _captum_available = False

_tensorboard_available = importlib.util.find_spec("tensorboard") is not None \
                         or importlib.util.find_spec("tensorboardX") is not None
try:
    from torch.utils.tensorboard import SummaryWriter  # pylint: disable=import-outside-toplevel

    trainer_log.info(SummaryWriter)
except (ImportError, AttributeError):
    _tensorboard_available = False


def is_captum_available():
    return _captum_available


def is_tensorboard_available():
    return _tensorboard_available


def is_matplotlib_available():
    return importlib.util.find_spec("matplotlib") is not None



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from towhee.engine.operator_runner.runner_base import RunnerBase
from towhee.errors import OpIOTypeError


class FilterRunner(RunnerBase):
    """
    FilterRunner

    Filter operator's output must be a bool type, if the output is true, pass data to
    the next operator else drop it.
    """

    def _set_outputs(self, output: bool):
        if not isinstance(output, bool):
            raise OpIOTypeError('Filter operator\'s output must be a bool type, this one is {}'.format(type(output)))

        if output:
            self._writer.write(self._row_data)
        else:
            self._frame.empty = True
            self._writer.write((self._frame, ))



from queue import Queue
import asyncio
import threading

from towhee.utils.log import engine_log
from towhee.functional.option import Option, Empty, _Reason
from towhee.functional.mixins.parallel import EOS


def _map_task_ray(unary_op):
    def map_wrapper(x):
        try:
            if isinstance(x, Option):
                return x.map(unary_op)
            else:
                return unary_op(x)
        except Exception as e:  # pylint: disable=broad-except
            engine_log.warning(f'{e}, please check {x} with op {unary_op}. Continue...')  # pylint: disable=logging-fstring-interpolation
            return Empty(_Reason(x, e))

    return map_wrapper


class RayMixin:
    """
    Mixin for parallel ray execution.
    """

    def _ray_pmap(self, unary_op, num_worker=None):
        import ray #pylint: disable=import-outside-toplevel
        ###### ACTOR: Currently much slower, may have future use.

        # @ray.remote
        # class RemoteActor:
        #     def remote_runner(self, val):
        #         return _map_task_ray(unary_op)(val)

        # pool = ray.util.ActorPool([RemoteActor.remote() for _ in range(num_worker)])

        # pool.submit(lambda a, v: a.remote_runner.remote(v), x)
        # if pool.has_next():
        #    queue.put(pool.get_next())

        ###### TASK: Need a good way of transferring the model.
        if num_worker is not None:
            pass
        elif self.get_executor() is not None:
            num_worker = self._num_worker
        else:
            num_worker = 2

        queue = Queue(num_worker)
        loop = asyncio.new_event_loop()

        def inner():
            while True:
                x = queue.get()
                if isinstance(x, EOS):
                    break
                else:
                    yield x

        @ray.remote
        def remote_runner(val):
            return _map_task_ray(unary_op)(val)

        async def worker():
            buff = []
            for x in self:
                if len(buff) == num_worker:
                    queue.put(await buff.pop(0))
                buff.append(asyncio.wrap_future(remote_runner.remote(x).future()))
            while len(buff) > 0:
                queue.put(await buff.pop(0))
            queue.put(EOS())

        def worker_wrapper():
            loop.run_until_complete(worker())
            loop.close()

        t = threading.Thread(target=worker_wrapper)
        t.start()

        return self._factory(inner())



# Copyright 2021 biubug6 . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.

#adapted from https://github.com/biubug6/Pytorch_Retinaface
from itertools import product
from typing import Dict, List
from math import ceil

import torch

class PriorBox(object):
    """
    PriorBox

    Anchor boxes utility for decoding the prior box.

    Args:
        cfg (`Dict`):
            retinaface configs.
        image_size (`Linst[int]`):
            input image size.
    """
    def __init__(self, cfg: Dict, image_size: List[int]=None):
        super().__init__()
        self.min_sizes = cfg['min_sizes']
        self.steps = cfg['steps']
        self.clip = cfg['clip']
        self.image_size = image_size
        self.feature_maps = [[ceil(self.image_size[0]/step), ceil(self.image_size[1]/step)] for step in self.steps]

    def forward(self):
        anchors = []
        for k, f in enumerate(self.feature_maps):
            min_sizes = self.min_sizes[k]
            for i, j in product(range(f[0]), range(f[1])):
                for min_size in min_sizes:
                    s_kx = min_size / self.image_size[1]
                    s_ky = min_size / self.image_size[0]
                    dense_cx = [x * self.steps[k] / self.image_size[1] for x in [j + 0.5]]
                    dense_cy = [y * self.steps[k] / self.image_size[0] for y in [i + 0.5]]
                    for cy, cx in product(dense_cy, dense_cx):
                        anchors += [cx, cy, s_kx, s_ky]

        # back to torch land
        output = torch.Tensor(anchors).view(-1, 4)
        if self.clip:
            output.clamp_(max=1, min=0)
        return output



# Copyright 2022 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional, List, Any, Tuple
from PIL import Image
from torchvision import transforms
from torchvision.io import read_image
from pathlib import Path
from torch import nn
from towhee.trainer.optimization.optimization import get_scheduler, get_warmup_steps
from towhee.trainer.utils.trainer_utils import _construct_scheduler_from_config
from towhee.trainer.training_config import TrainingConfig
from towhee.trainer.utils.file_utils import is_captum_available, is_matplotlib_available
from towhee.utils.log import trainer_log

import torch
import torch.nn.functional as F
import numpy as np
import random
import os
import copy


def _show_grid(row_grid_list, cls_list):
    if not is_matplotlib_available():
        trainer_log.warning('Matplotlib is not available.')
    import matplotlib.pylab as plt  # pylint: disable=import-outside-toplevel

    _, axs = plt.subplots(nrows=len(row_grid_list), ncols=len(row_grid_list[0]), squeeze=False, figsize=(13, 13))
    for row, (row_gird, _) in enumerate(zip(row_grid_list, cls_list)):
        for col, img in enumerate(row_gird):
            show_img = img.numpy().transpose(1, 2, 0)
            axs[row, col].imshow(show_img)
            axs[row, col].set(xticklabels=[], yticklabels=[], xticks=[], yticks=[])
    plt.show()


def image_folder_sample_show(root: str, rows: int = 4, cols: int = 4, img_size: int = 224,
                             classes: Optional[List[str]] = None):
    """
    Show sample images for `torchvision.datasets.ImageFolder`,
    Put another way, the image root path structure just need to be like this:
    ```
    root-|
        #classname1#-|
            image-1.jpg
            image-2.jpg
        #classname2-|
            image-1.jpg
            image-2.jpg
        |
        |
        #classnameN-|
            image-1.jpg
            image-2.jpg
    ```
    it's useful to observe the images for different categories.
    You can randomly sample categories, or specify show classes.
    Args:
        root (`str`):
            Root of `ImageFolder` instance.
        rows (`int`):
            Count of show rows.
        cols (`int`):
            Count of show cols.
        img_size (`int` or `tuple`):
            Show image shape, if int, it's a square.
        classes (`Optional[list[str]]`):
            The specified classes you want to see. If not None, param `rows` will be invalid.

    """
    transform = transforms.Compose([
        transforms.Resize(img_size),
        transforms.CenterCrop(img_size),
    ])
    root_path = Path(root)
    rows = min(len(list(root_path.iterdir())), rows)
    if classes:
        sample_class_list = [root_path / cls for cls in classes]
    else:
        sample_class_list = random.sample(list(root_path.iterdir()), rows)
    row_grid_list = []
    cls_list = []
    for class_path in sample_class_list:
        cls = class_path.name
        class_sample_num = min(len(list(class_path.iterdir())), cols)
        class_sample_img_list = random.sample(
            list(class_path.iterdir()),
            class_sample_num
        )
        trans_img_list = []
        for img_path in class_sample_img_list:
            img = read_image(str(img_path))
            trans_img = transform(img)
            trans_img_list.append(trans_img)
        row_grid_list.append(trans_img_list)
        cls_list.append(cls)
    _show_grid(row_grid_list, cls_list)


def image_folder_statistic(root: str, classes: Optional[List[str]] = None, show_bar: Optional[bool] = False):
    """
    Show count statistic infos for `torchvision.datasets.ImageFolder`.
    Put another way, the image root path structure just need to be like this:
    ```
    root-|
        #classname1#-|
            image-1.jpg
            image-2.jpg
        #classname2-|
            image-1.jpg
            image-2.jpg
        |
        |
        #classnameN-|
            image-1.jpg
            image-2.jpg
    ```
    Args:
        root (`str`):
            Root of `ImageFolder` instance.
        classes (`Optional[List[str]]`):
            The specified classes of statistic. If not None, all class statistic will be return.
        show_bar (`Optional[bool]`):
            Whether showing bar graph.

    Returns:
        (`dict`)
            Count statistic info, where key is category name, value is the count.
    """
    if not is_matplotlib_available():
        trainer_log.warning('Matplotlib is not available.')
    import matplotlib.pylab as plt  # pylint: disable=import-outside-toplevel

    root_path = Path(root)
    if classes:
        class_list = [root_path / cls for cls in classes]
    else:
        class_list = os.listdir(root)
    count_dict = {}
    for cls in class_list:
        cls_num = len(list((root_path / cls).iterdir()))
        count_dict[cls] = cls_num
    if show_bar:
        plt.figure(figsize=(30, 10))
        plt.bar(x=count_dict.keys(), height=count_dict.values(), color='red')
        plt.xticks(rotation=90)
        plt.tight_layout()
        plt.grid(True)
        plt.title(root_path.name)
        plt.show()
    return count_dict


def show_transform(image_path: str, transform: Any, sample_num: int = 6):
    """
    Show the result which `torchvision.transforms` or any other callable function act on the image.
    Only the Image transforms is supported. Such as `ToTensor` or `Normalize` is invalid.
    Args:
        image_path (`str`):
            The original image path.
        transform (`Any`):
            torchvision.transforms` or any other callable function.
        sample_num (`int`):
            Sample number.

    """
    if not is_matplotlib_available():
        trainer_log.warning('Matplotlib is not available.')
    import matplotlib.pylab as plt  # pylint: disable=import-outside-toplevel
    plt.rcParams['savefig.bbox'] = 'tight'
    orig_img = Image.open(image_path)
    trans_img_list = [transform(orig_img) for _ in range(sample_num)]
    _plot_transform(orig_img, trans_img_list)


def _plot_transform(orig_img, imgs, with_orig=True, row_title=None, **imshow_kwargs):
    if not is_matplotlib_available():
        trainer_log.warning('Matplotlib is not available.')
    import matplotlib.pylab as plt  # pylint: disable=import-outside-toplevel
    if not isinstance(imgs[0], list):
        imgs = [imgs]
    num_rows = len(imgs)
    num_cols = len(imgs[0]) + with_orig
    _, axs = plt.subplots(nrows=num_rows, ncols=num_cols, squeeze=False, figsize=(12, 12))
    for row_idx, row in enumerate(imgs):
        row = [orig_img] + row if with_orig else row
        for col_idx, img in enumerate(row):
            ax = axs[row_idx, col_idx]
            ax.imshow(np.asarray(img), **imshow_kwargs)
            ax.set(xticklabels=[], yticklabels=[], xticks=[], yticks=[])
    if with_orig:
        axs[0, 0].set(title='Original image')
        axs[0, 0].title.set_size(8)
    if row_title is not None:
        for row_idx in range(num_rows):
            axs[row_idx, 0].set(ylabel=row_title[row_idx])

    plt.tight_layout()
    plt.show()


def plot_lrs_for_scheduler(optimizer: torch.optim.Optimizer, scheduler: '_LRScheduler', total_steps: int):
    """
    Plot the lr curve for a specified optimizer and scheduler instance.
    Args:
        optimizer (`Optimizer`):
            The optimizer contains the training parameters of a model.
        scheduler (`_LRScheduler`):
            The lr scheduler for an optimizer.
        total_steps (`int`):
            Total training step number.

    """
    if not is_matplotlib_available():
        trainer_log.warning('Matplotlib is not available.')
    import matplotlib.pylab as plt  # pylint: disable=import-outside-toplevel
    lrs = []
    for _ in range(total_steps):
        optimizer.step()
        lrs.append(optimizer.param_groups[0]['lr'])
        scheduler.step()
    plt.plot(range(total_steps), lrs)
    plt.show()


def _create_scheduler_from_config(configs: TrainingConfig, num_training_steps: int,
                                  optimizer: torch.optim.Optimizer = None):
    assert isinstance(configs.lr_scheduler_type, (str, dict))
    lr_scheduler = None
    if isinstance(configs.lr_scheduler_type, str):
        lr_scheduler = get_scheduler(
            configs.lr_scheduler_type,
            optimizer=optimizer if optimizer is None else optimizer,
            num_warmup_steps=get_warmup_steps(num_training_steps, configs.warmup_steps, configs.warmup_ratio),
            num_training_steps=num_training_steps,
        )
    elif isinstance(configs.lr_scheduler_type, dict):
        lr_scheduler = _construct_scheduler_from_config(optimizer,
                                                        torch.optim.lr_scheduler,
                                                        configs.lr_scheduler_type)
    return lr_scheduler


def plot_lrs_for_config(configs: TrainingConfig, num_training_steps: int, start_lr: int = 100):
    """
    Plot the lr for a config.
    It's useful to visualize how the lr scheduler control the lr by the yaml file configs.
    Args:
        configs (`TrainingConfig`):
            The training config, which can be load from the yaml file.
        num_training_steps (`int`):
            Total training step number.
        start_lr (`int`):
            The start lr value.

    """
    model = nn.Linear(2, 1)
    optimizer = torch.optim.SGD(model.parameters(), lr=start_lr)
    lr_scheduler = _create_scheduler_from_config(configs, num_training_steps, optimizer)
    plot_lrs_for_scheduler(optimizer, lr_scheduler, num_training_steps)


def predict_image_classification(model: nn.Module, input_: torch.Tensor):
    """
    Predict using an image classification model.
    Args:
        model (`nn.Module`):
            Pytorch model.
        input_ (`Tensor`):
            Input image tensor.
    Returns:
        (`tuple`)
            Prediction score which max is 1, and label idx.

    """
    output = model(input_)
    output = F.softmax(output, dim=1)
    prediction_score, pred_label_idx = torch.topk(output, 1)
    if isinstance(pred_label_idx, torch.Tensor):
        pred_label_idx = pred_label_idx.squeeze().item()
    prediction_score = prediction_score.squeeze().detach().item()
    return prediction_score, pred_label_idx


def interpret_image_classification(model: nn.Module, image: Any, eval_transform: transforms.Compose, method: str,
                                   fig_size: Tuple = (10, 10), cmap: Any = 'OrRd', pred_label_idx: int = None,
                                   titles: List = None, **kwargs: Any):
    """
    Use Captum to interpret the specified class of network output.
    Captum should be installed.
    Args:
        model (`nn.Module`):
            Pytorch module.
        image (`Any`):
            The image before do transform.
            It can be produced by either pytorch `DataLoader` or read by `Image.open()` using PIL.
        eval_transform (`transforms.Compose`):
            Evaluation transform.
        method (`method`):
            It can be in ['Occlusion', 'IntegratedGradients', 'GradientShap', 'Saliency'].
        fig_size (`Tuple`):
            Figure plotting size.
        cmap (`Any`):
            Matplotlib colormap.
        pred_label_idx (`int`):
            If None, use the predicted class automatically.
        titles (`List`):
            Plotted titles of the two axs in the figure.
        **kwargs (`Any`):
            Keyword Args.
    Returns:
        (`tuple`)
            Prediction score and label idx. If input label is specified, the prediction score is None.
    """
    if not is_captum_available():
        trainer_log.warning('You should install Captum first. Please run `pip install captum`.')
        return None
    from captum.attr import visualization as viz  # pylint: disable=import-outside-toplevel
    assert method in ['Occlusion', 'IntegratedGradients', 'GradientShap', 'Saliency']

    def _viz_image_attr_multiple(attribution, trans_image: torch.Tensor, fig_size, cmap,
                                 titles: List[str] = None, **kwargs):
        _ = viz.visualize_image_attr_multiple(
            np.transpose(attribution.squeeze().cpu().detach().numpy(), (1, 2, 0)),
            np.transpose(trans_image.squeeze().cpu().detach().numpy(), (1, 2, 0)),
            ['original_image', 'blended_heat_map'],
            ['all', 'positive'],
            show_colorbar=True,
            outlier_perc=2,
            fig_size=fig_size,
            cmap=cmap,
            titles=titles,
            **kwargs
        )

    model.eval()

    #  prepare image and input
    input_ = eval_transform(image)
    trans_without_norm = copy.deepcopy(eval_transform)
    for trans in trans_without_norm.transforms:
        if trans.__class__.__name__ == 'Normalize':
            trans_without_norm.transforms.remove(trans)
    trans_image = trans_without_norm(image)
    if len(trans_image.shape) == 3:
        trans_image = trans_image.unsqueeze(0)
    if len(input_.shape) == 3:
        input_ = input_.unsqueeze(0)

    prediction_score = None
    if pred_label_idx is None:
        prediction_score, pred_label_idx = predict_image_classification(model, input_)

    if titles is None:
        titles = ['Origin Image', 'Method:' + method]

    if method == 'Occlusion':
        from captum.attr import Occlusion  # pylint: disable=import-outside-toplevel
        occlusion = Occlusion(model)

        attributions_occ = occlusion.attribute(input_,
                                               strides=(3, 8, 8),
                                               target=pred_label_idx,
                                               sliding_window_shapes=(3, 15, 15),
                                               baselines=0)
        _viz_image_attr_multiple(attributions_occ, trans_image, fig_size, cmap, titles=titles, **kwargs)

    if method == 'IntegratedGradients':
        from captum.attr import NoiseTunnel  # pylint: disable=import-outside-toplevel
        from captum.attr import IntegratedGradients  # pylint: disable=import-outside-toplevel
        integrated_gradients = IntegratedGradients(model)
        noise_tunnel = NoiseTunnel(integrated_gradients)

        attributions_ig_nt = noise_tunnel.attribute(input_, nt_samples=10, nt_type='smoothgrad_sq',
                                                    target=pred_label_idx)
        _viz_image_attr_multiple(attributions_ig_nt, trans_image, fig_size, cmap, titles=titles, **kwargs)

    if method == 'GradientShap':
        from captum.attr import GradientShap, NoiseTunnel  # pylint: disable=import-outside-toplevel
        gradient_shap = GradientShap(model)
        rand_img_dist = torch.cat([input_ * 0, input_ * 1])
        attributions_gs = gradient_shap.attribute(input_,
                                                  n_samples=50,
                                                  stdevs=0.0001,
                                                  baselines=rand_img_dist,
                                                  target=pred_label_idx)
        _viz_image_attr_multiple(attributions_gs, trans_image, fig_size, cmap, titles=titles, **kwargs)

    if method == 'Saliency':
        from captum.attr import Saliency  # pylint: disable=import-outside-toplevel
        saliency = Saliency(model)
        grads = saliency.attribute(input_, target=pred_label_idx)
        _viz_image_attr_multiple(grads, trans_image, fig_size, cmap, titles=titles, **kwargs)

    return prediction_score, pred_label_idx



# Copyright 2021 Ross Wightman . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.

""" DropPath

PyTorch implementations of DropPath (Stochastic Depth) regularization layers.

Papers:
Deep Networks with Stochastic Depth (https://arxiv.org/abs/1603.09382)

Copyright 2020 Ross Wightman
"""

import torch
from torch import nn

def drop_path(x, drop_prob: float = 0., training: bool = False):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).

    This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
    changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
    'survival rate' as the argument.

    """
    if drop_prob == 0. or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (x.ndim - 1)  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = keep_prob + torch.rand(shape, dtype=x.dtype, device=x.device)
    random_tensor.floor_()  # binarize
    output = x.div(keep_prob) * random_tensor
    return output


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks).
    """
    def __init__(self, drop_prob=None):
        super().__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training)
